from __future__ import annotations
from bcps import core


class Catamin:
    def __init__(self, amount: int):
        self.amount = amount

    @staticmethod
    def read(stream: core.Data) -> Catamin:
        amount = stream.read_int()
        return Catamin(amount)

    def write(self, stream: core.Data):
        stream.write_int(self.amount)

    def serialize(self) -> int:
        return self.amount

    @staticmethod
    def deserialize(data: int) -> Catamin:
        return Catamin(data)

    def __repr__(self):
        return f"Catamin({self.amount})"

    def __str__(self):
        return f"Catamin({self.amount})"


class Catamins:
    def __init__(self, catamins: list[Catamin]):
        self.catamins = catamins

    @staticmethod
    def read(stream: core.Data) -> Catamins:
        total = stream.read_int()
        catamins: list[Catamin] = []
        for _ in range(total):
            catamins.append(Catamin.read(stream))
        return Catamins(catamins)

    def write(self, stream: core.Data):
        stream.write_int(len(self.catamins))
        for catamin in self.catamins:
            catamin.write(stream)

    def serialize(self) -> list[int]:
        return [catamin.serialize() for catamin in self.catamins]

    @staticmethod
    def deserialize(data: list[int]) -> Catamins:
        return Catamins([Catamin.deserialize(catamin) for catamin in data])

    def __repr__(self):
        return f"Catamins({self.catamins})"

    def __str__(self):
        return f"Catamins({self.catamins})"



from __future__ import annotations
from typing import Any

from bcps import core
from bcps.cli import color, dialog_creator


class Mission:
    def __init__(
        self,
        clear_state: int | None = None,
        requirement: int | None = None,
        progress_type: int | None = None,
        gamatoto_value: int | None = None,
        nyancombo_value: int | None = None,
        user_rank_value: int | None = None,
        expiry_value: int | None = None,
        preparing_value: int | bool | None = None,
    ):
        self.clear_state = clear_state
        self.requirement = requirement
        self.progress_type = progress_type
        self.gamatoto_value = gamatoto_value
        self.nyancombo_value = nyancombo_value
        self.user_rank_value = user_rank_value
        self.expiry_value = expiry_value
        self.preparing_value = preparing_value

    @staticmethod
    def init() -> Mission:
        return Mission(
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
        )

    def serialize(self) -> dict[str, Any]:
        return {
            "clear_state": self.clear_state,
            "requirement": self.requirement,
            "progress_type": self.progress_type,
            "gamatoto_value": self.gamatoto_value,
            "nyancombo_value": self.nyancombo_value,
            "user_rank_value": self.user_rank_value,
            "expiry_value": self.expiry_value,
            "preparing_value": self.preparing_value,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Mission:
        return Mission(
            data["clear_state"],
            data["requirement"],
            data["progress_type"],
            data["gamatoto_value"],
            data["nyancombo_value"],
            data["user_rank_value"],
            data["expiry_value"],
            data["preparing_value"],
        )

    def __repr__(self):
        return f"Mission({self.clear_state}, {self.requirement}, {self.progress_type}, {self.gamatoto_value}, {self.nyancombo_value}, {self.user_rank_value}, {self.expiry_value}, {self.preparing_value})"

    def __str__(self):
        return self.__repr__()


class Missions:
    def __init__(
        self,
        clear_states: dict[int, int],
        requirements: dict[int, int],
        progress_types: dict[int, int],
        gamatoto_values: dict[int, int],
        nyancombo_values: dict[int, int],
        user_rank_values: dict[int, int],
        expiry_values: dict[int, int],
        preparing_values: dict[int, int | bool],
    ):
        self.clear_states = clear_states
        self.requirements = requirements
        self.progress_types = progress_types
        self.gamatoto_values = gamatoto_values
        self.nyancombo_values = nyancombo_values
        self.user_rank_values = user_rank_values
        self.expiry_values = expiry_values
        self.preparing_values = preparing_values
        self.weekly_missions: dict[int, bool] = {}

    @staticmethod
    def init() -> Missions:
        return Missions(
            {},
            {},
            {},
            {},
            {},
            {},
            {},
            {},
        )

    @staticmethod
    def read(stream: core.Data, gv: core.GameVersion) -> Missions:
        clear_states: dict[int, int] = stream.read_int_int_dict()
        requirements: dict[int, int] = stream.read_int_int_dict()
        progress_types: dict[int, int] = stream.read_int_int_dict()
        gamatoto_values: dict[int, int] = stream.read_int_int_dict()
        nyancombo_values: dict[int, int] = stream.read_int_int_dict()
        user_rank_values: dict[int, int] = stream.read_int_int_dict()
        expiry_values: dict[int, int] = stream.read_int_int_dict()
        preparing_values: dict[int, int | bool] = {}

        for _ in range(stream.read_int()):
            key = stream.read_int()
            if gv < 90300:
                preparing_values[key] = stream.read_bool()
            else:
                preparing_values[key] = stream.read_int()

        return Missions(
            clear_states,
            requirements,
            progress_types,
            gamatoto_values,
            nyancombo_values,
            user_rank_values,
            expiry_values,
            preparing_values,
        )

    def write(self, stream: core.Data, gv: core.GameVersion):
        stream.write_int_int_dict(self.clear_states)
        stream.write_int_int_dict(self.requirements)
        stream.write_int_int_dict(self.progress_types)
        stream.write_int_int_dict(self.gamatoto_values)
        stream.write_int_int_dict(self.nyancombo_values)
        stream.write_int_int_dict(self.user_rank_values)
        stream.write_int_int_dict(self.expiry_values)

        stream.write_int(len(self.preparing_values))
        for key, value in self.preparing_values.items():
            stream.write_int(key)
            if gv < 90300:
                stream.write_bool(bool(value))
            else:
                stream.write_int(int(value))

    def read_weekly_missions(self, stream: core.Data):
        self.weekly_missions: dict[int, bool] = {}
        for _ in range(stream.read_int()):
            key = stream.read_int()
            self.weekly_missions[key] = stream.read_bool()

    def write_weekly_missions(self, stream: core.Data):
        stream.write_int(len(self.weekly_missions))
        for key, value in self.weekly_missions.items():
            stream.write_int(key)
            stream.write_bool(value)

    def serialize(self) -> dict[str, Any]:
        return {
            "clear_states": self.clear_states,
            "requirements": self.requirements,
            "progress_types": self.progress_types,
            "gamatoto_values": self.gamatoto_values,
            "nyancombo_values": self.nyancombo_values,
            "user_rank_values": self.user_rank_values,
            "expiry_values": self.expiry_values,
            "preparing_values": self.preparing_values,
            "weekly_missions": self.weekly_missions,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]):
        missions = Missions(
            data.get("clear_states", {}),
            data.get("requirements", {}),
            data.get("progress_types", {}),
            data.get("gamatoto_values", {}),
            data.get("nyancombo_values", {}),
            data.get("user_rank_values", {}),
            data.get("expiry_values", {}),
            data.get("preparing_values", {}),
        )
        missions.weekly_missions = data.get("weekly_missions", {})
        return missions

    def __repr__(self):
        return f"<Missions {self.serialize()}>"

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def edit_missions(save_file: core.SaveFile):
        missions = save_file.missions

        names = core.core_data.get_mission_names(save_file)
        conditions = core.core_data.get_mission_conditions(save_file)
        if names.names is None or conditions.conditions is None:
            return
        options: list[str] = []
        mssion_ids: list[int] = []
        for mission_id, name in names.names.items():
            if mission_id in missions.clear_states:
                name = name.split("<br>")[0]
                condition = conditions.conditions.get(mission_id)
                if not condition:
                    continue
                name = name.replace("%d", str(condition.progress_count))
                if "%@" in name and len(condition.conditions_value) > 2:
                    name = name.replace("%@", str(condition.conditions_value[2]))
                options.append(name)
                mssion_ids.append(mission_id)

        re_claim = dialog_creator.ChoiceInput.from_reduced(
            ["complete_reward", "complete_claim", "uncomplete"],
            dialog="select_mission_claim",
            single_choice=True,
        ).single_choice()
        if re_claim is None:
            return
        re_claim -= 1

        choices, _ = dialog_creator.ChoiceInput.from_reduced(
            options, dialog="select_missions"
        ).multiple_choice(localized_options=False)
        if choices is None:
            return
        for choice in choices:
            mission_id = mssion_ids[choice]
            if re_claim == 0:
                missions.clear_states[mission_id] = 2
                condition = conditions.get_condition(mission_id)
                if condition is not None:
                    missions.requirements[mission_id] = condition.progress_count
            elif re_claim == 1:
                missions.clear_states[mission_id] = 4
                condition = conditions.get_condition(mission_id)
                if condition is not None:
                    missions.requirements[mission_id] = condition.progress_count
            elif re_claim == 2:
                missions.clear_states[mission_id] = 0
                if mission_id in missions.requirements:
                    missions.requirements[mission_id] = 0

        color.ColoredText.localize("missions_edited")


class MissionCondition:
    def __init__(
        self,
        mission_id: int,
        mission_type: int,
        conditions_type: int,
        progress_count: int,
        conditions_value: list[int],
    ):
        self.mission_id = mission_id
        self.mission_type = mission_type
        self.conditions_type = conditions_type
        self.progress_count = progress_count
        self.conditions_value = conditions_value


class MissionConditions:
    def __init__(self, save: core.SaveFile):
        self.save = save
        self.conditions = self.get_conditions()

    def get_conditions(self) -> dict[int, MissionCondition] | None:
        file_name = "Mission_Condition.csv"
        gdg = core.core_data.get_game_data_getter(self.save)
        file = gdg.download("DataLocal", file_name)
        if file is None:
            return None
        csv = core.CSV(file)
        conditions: dict[int, MissionCondition] = {}
        for row in csv:
            conditions[row[0].to_int()] = MissionCondition(
                row[0].to_int(),
                row[1].to_int(),
                row[2].to_int(),
                row[3].to_int(),
                row[4:].to_int_list(),
            )
        return conditions

    def get_condition(self, mission_id: int) -> MissionCondition | None:
        if self.conditions is None:
            return None
        return self.conditions.get(mission_id)


class MissionNames:
    def __init__(self, save: core.SaveFile):
        self.save = save
        self.names = self.get_names()

    def get_names(self) -> dict[int, str] | None:
        file_name = "Mission_Name.csv"
        gdg = core.core_data.get_game_data_getter(self.save)
        file = gdg.download("resLocal", file_name)
        if file is None:
            return None
        csv = core.CSV(
            file, delimiter=core.Delimeter.from_country_code_res(self.save.cc)
        )
        names: dict[int, str] = {}
        for row in csv:
            names[row[0].to_int()] = row[1].to_str()
        return names



from __future__ import annotations
import datetime
import random
import time
from typing import Any

from bcps import core
from bcps.cli import dialog_creator, color


class NyankoClub:
    def __init__(
        self,
        officer_id: int,
        total_renewal_times: int,
        start_date_now: float,
        end_date_now: float,
        start_date_next: float,
        end_date_next: float,
        start_date_total: float,
        end_date_total: float,
        time_error_end: float,
        total_state_updates: int,
        login_bonus_date: float,
        claimed_rewards: dict[int, int],
        remaing_days_popup: float,
        first_popup_flag: bool,
        badge_flag: bool | None = None,
    ):
        self.officer_id = officer_id
        self.total_renewal_times = total_renewal_times
        self.start_date_now = start_date_now
        self.end_date_now = end_date_now
        self.start_date_next = start_date_next
        self.end_date_next = end_date_next
        self.start_date_total = start_date_total
        self.end_date_total = end_date_total
        self.time_error_end = time_error_end
        self.total_state_updates = total_state_updates
        self.login_bonus_date = login_bonus_date
        self.claimed_rewards = claimed_rewards
        self.remaing_days_popup = remaing_days_popup
        self.first_popup_flag = first_popup_flag
        self.badge_flag = badge_flag

    @staticmethod
    def init() -> NyankoClub:
        return NyankoClub(
            0,
            0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0,
            0.0,
            {},
            0.0,
            False,
            False,
        )

    @staticmethod
    def read(data: core.Data, gv: core.GameVersion) -> NyankoClub:
        officer_id = data.read_int()
        total_renewal_times = data.read_int()
        start_date_now = data.read_double()
        end_date_now = data.read_double()
        start_date_next = data.read_double()
        end_date_next = data.read_double()
        start_date_total = data.read_double()
        end_date_total = data.read_double()
        time_error_end = data.read_double()
        total_state_updates = data.read_int()
        login_bonus_date = data.read_double()
        claimed_rewards = data.read_int_int_dict()
        remaing_days_popup = data.read_double()
        first_popup_flag = data.read_bool()
        if gv >= 80100:
            badge_flag = data.read_bool()
        else:
            badge_flag = None
        return NyankoClub(
            officer_id,
            total_renewal_times,
            start_date_now,
            end_date_now,
            start_date_next,
            end_date_next,
            start_date_total,
            end_date_total,
            time_error_end,
            total_state_updates,
            login_bonus_date,
            claimed_rewards,
            remaing_days_popup,
            first_popup_flag,
            badge_flag,
        )

    def write(self, data: core.Data, gv: core.GameVersion):
        data.write_int(self.officer_id)
        data.write_int(self.total_renewal_times)
        data.write_double(self.start_date_now)
        data.write_double(self.end_date_now)
        data.write_double(self.start_date_next)
        data.write_double(self.end_date_next)
        data.write_double(self.start_date_total)
        data.write_double(self.end_date_total)
        data.write_double(self.time_error_end)
        data.write_int(self.total_state_updates)
        data.write_double(self.login_bonus_date)
        data.write_int_int_dict(self.claimed_rewards)
        data.write_double(self.remaing_days_popup)
        data.write_bool(self.first_popup_flag)
        if gv >= 80100:
            data.write_bool(self.badge_flag or False)

    def serialize(self) -> dict[str, Any]:
        return {
            "officer_id": self.officer_id,
            "total_renewal_times": self.total_renewal_times,
            "start_date_now": self.start_date_now,
            "end_date_now": self.end_date_now,
            "start_date_next": self.start_date_next,
            "end_date_next": self.end_date_next,
            "start_date_total": self.start_date_total,
            "end_date_total": self.end_date_total,
            "time_error_end": self.time_error_end,
            "total_state_updates": self.total_state_updates,
            "login_bonus_date": self.login_bonus_date,
            "claimed_rewards": self.claimed_rewards,
            "remaing_days_popup": self.remaing_days_popup,
            "first_popup_flag": self.first_popup_flag,
            "badge_flag": self.badge_flag,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> NyankoClub:
        return NyankoClub(
            data.get("officer_id", 0),
            data.get("total_renewal_times", 0),
            data.get("start_date_now", 0.0),
            data.get("end_date_now", 0.0),
            data.get("start_date_next", 0.0),
            data.get("end_date_next", 0.0),
            data.get("start_date_total", 0.0),
            data.get("end_date_total", 0.0),
            data.get("time_error_end", 0.0),
            data.get("total_state_updates", 0),
            data.get("login_bonus_date", 0.0),
            data.get("claimed_rewards", {}),
            data.get("remaing_days_popup", 0.0),
            data.get("first_popup_flag", False),
            data.get("badge_flag", False),
        )

    def __repr__(self):
        return f"<NyankoClub {self.officer_id}>"

    def __str__(self):
        return f"NyankoClub {self.officer_id}"

    def get_gold_pass(self, officer_id: int, total_days: int, save_file: core.SaveFile):
        self.officer_id = officer_id
        start_date_now = int(time.time())
        end_date_now = (
            start_date_now + datetime.timedelta(days=total_days).total_seconds()
        )
        end_date_total = (
            start_date_now + datetime.timedelta(days=total_days * 2).total_seconds()
        )

        self.total_renewal_times = 2
        self.start_date_now = start_date_now
        self.end_date_now = end_date_now

        self.start_date_next = end_date_now
        self.end_date_next = end_date_total

        self.start_date_total = start_date_now
        self.end_date_total = end_date_total

        self.time_error_end = start_date_now

        self.total_state_updates = 2

        self.login_bonus_date = end_date_now

        self.remaing_days_popup = 0.0
        self.first_popup_flag = True
        self.badge_flag = False

        login = save_file.logins.get_login(5100)
        if login is not None:
            login.count = 0

        self.claimed_rewards = {}

    def remove_gold_pass(self, save_file: core.SaveFile):
        self.officer_id = -1
        self.total_renewal_times = 0
        self.start_date_now = 0.0
        self.end_date_now = 0.0
        self.start_date_next = 0.0
        self.end_date_next = 0.0
        self.start_date_total = 0.0
        self.end_date_total = 0.0
        self.time_error_end = 0.0
        self.total_state_updates = 0
        self.login_bonus_date = 0.0
        self.remaing_days_popup = 0.0
        self.first_popup_flag = False
        self.badge_flag = False

        login = save_file.logins.get_login(5100)
        if login is not None:
            login.count = 0

        self.claimed_rewards = {}

    @staticmethod
    def get_random_officer_id() -> int:
        return random.randint(1, 2**16 - 1)

    @staticmethod
    def edit_gold_pass(save_file: core.SaveFile):
        club = save_file.officer_pass.gold_pass

        officer_id = color.ColoredInput().localize("gold_pass_dialog").strip()
        if not officer_id:
            officer_id = NyankoClub.get_random_officer_id()

        if officer_id == "-1":
            officer_id = -1
        else:
            try:
                officer_id = int(officer_id)
            except ValueError:
                officer_id = NyankoClub.get_random_officer_id()
            officer_id = dialog_creator.IntInput().clamp_value(officer_id)

        if officer_id == -1:
            club.remove_gold_pass(save_file)
            color.ColoredText.localize("gold_pass_remove_success")
        else:
            club.get_gold_pass(officer_id, 30, save_file)
            color.ColoredText.localize("gold_pass_get_success", id=officer_id)



from __future__ import annotations
from dataclasses import dataclass
from typing import Any
from bcps import core
from bcps.cli import dialog_creator, color


@dataclass
class LevelPartRecipeUnlock:
    index: int
    cannon_id: int
    part_id: int
    unknown: int
    unknown2: int
    level: int


class CastleRecipeUnlock:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.level_part_recipe_unlocks = self.get_recipe_unlocks()

    def get_recipe_unlocks(self) -> list[LevelPartRecipeUnlock] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("DataLocal", "CastleRecipeUnlock.csv")
        if data is None:
            return None
        csv = core.CSV(data)
        level_part_recipe_unlocks: list[LevelPartRecipeUnlock] = []
        for i, line in enumerate(csv):
            level_part_recipe_unlocks.append(
                LevelPartRecipeUnlock(
                    index=i,
                    cannon_id=line[0].to_int(),
                    part_id=line[1].to_int(),
                    unknown=line[2].to_int(),
                    unknown2=line[3].to_int(),
                    level=line[4].to_int(),
                )
            )

        return level_part_recipe_unlocks

    def get_recipe_unlock(self, index: int) -> LevelPartRecipeUnlock | None:
        if self.level_part_recipe_unlocks is None:
            return None
        for recipe_unlock in self.level_part_recipe_unlocks:
            if recipe_unlock.index == index:
                return recipe_unlock

        return None

    def get_max_level(self, cannon_id: int, part_id: int) -> int | None:
        if self.level_part_recipe_unlocks is None:
            return None
        max_level = 0

        for recipe_unlock in self.level_part_recipe_unlocks:
            if (
                recipe_unlock.cannon_id == cannon_id
                and recipe_unlock.part_id == part_id
            ):
                if recipe_unlock.level > max_level:
                    max_level = recipe_unlock.level

        return max_level

    def get_max_part_level(self, part_id: int) -> int | None:
        if self.level_part_recipe_unlocks is None:
            return None
        max_level = 0
        for recipe_unlock in self.level_part_recipe_unlocks:
            if recipe_unlock.part_id == part_id:
                if recipe_unlock.level > max_level:
                    max_level = recipe_unlock.level

        return max_level


@dataclass
class CannonDescription:
    cannon_id: int
    build_name: str
    foundation_build_description: str
    style_build_description: str
    effect_build_description: str
    cannon_build_description: str
    cannon_name: str
    foundation_name: str
    style_name: str
    effect_description: str
    improve_foundation_description: str
    improve_style_description: str
    improved_foundation_name: str
    improved_style_name: str
    improved_effect1_description: str
    improved_effect2_description: str

    def get_part_names(self) -> list[str]:
        effect_name = self.effect_build_description.split("<br>")[0]
        if not effect_name:
            effect_name = self.build_name
        return [
            effect_name,
            self.improve_foundation_description.split("<br>")[0],
            self.improve_style_description.split("<br>")[0],
        ]

    def get_part_name(self, index: int) -> str:
        return self.get_part_names()[index]

    def get_longest_part_name(self) -> str:
        return max(self.get_part_names(), key=len)

    def get_cannon_name(self) -> str:
        return self.cannon_name


class CannonDescriptions:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.cannon_descriptions = self.get_cannon_descriptions()

    def get_cannon_descriptions(self) -> list[CannonDescription] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("resLocal", "CastleRecipeDescriptions.csv")
        if data is None:
            return None
        csv = core.CSV(
            data,
            delimiter=core.Delimeter.from_country_code_res(self.save_file.cc),
            remove_empty=False,
        )
        cannon_descriptions: list[CannonDescription] = []
        for line in csv:
            cannon_descriptions.append(
                CannonDescription(
                    cannon_id=line[0].to_int(),
                    build_name=line[1].to_str(),
                    foundation_build_description=line[2].to_str(),
                    style_build_description=line[3].to_str(),
                    effect_build_description=line[4].to_str(),
                    cannon_build_description=line[5].to_str(),
                    cannon_name=line[6].to_str(),
                    foundation_name=line[7].to_str(),
                    style_name=line[8].to_str(),
                    effect_description=line[9].to_str(),
                    improve_foundation_description=line[10].to_str(),
                    improve_style_description=line[11].to_str(),
                    improved_foundation_name=line[12].to_str(),
                    improved_style_name=line[13].to_str(),
                    improved_effect1_description=line[14].to_str(),
                    improved_effect2_description=line[15].to_str(),
                )
            )

        return cannon_descriptions

    def get_cannon_description(self, cannon_id: int) -> CannonDescription | None:
        if self.cannon_descriptions is None:
            return None
        for cannon_description in self.cannon_descriptions:
            if cannon_description.cannon_id == cannon_id:
                return cannon_description

        return None

    def get_longest_longest_part_name(self) -> str | None:
        if self.cannon_descriptions is None:
            return None
        longest_part_name = ""
        for cannon_description in self.cannon_descriptions:
            l_name = cannon_description.get_longest_part_name()
            if len(l_name) > len(longest_part_name):
                longest_part_name = l_name

        return longest_part_name


class Cannon:
    def __init__(self, development: int, levels: list[int]):
        self.development = development
        self.levels = levels

    @staticmethod
    def init() -> Cannon:
        return Cannon(0, [])

    @staticmethod
    def read(stream: core.Data) -> Cannon:
        total = stream.read_int()
        levels: list[int] = []
        development = stream.read_int()
        for _ in range(total - 1):
            levels.append(stream.read_int())
        return Cannon(development, levels)

    def write(self, stream: core.Data):
        stream.write_int(len(self.levels) + 1)
        stream.write_int(self.development)
        for level in self.levels:
            stream.write_int(level)

    def serialize(self) -> list[int]:
        return [self.development] + self.levels

    @staticmethod
    def deserialize(data: list[int]) -> Cannon:
        return Cannon(data[0], data[1:])

    def __repr__(self):
        return f"Cannon({self.development}, {self.levels})"

    def __str__(self):
        return f"Cannon({self.development}, {self.levels})"


class Cannons:
    def __init__(self, cannons: dict[int, Cannon], selected_parts: list[list[int]]):
        self.cannons = cannons
        self.selected_parts = selected_parts

    @staticmethod
    def init(gv: core.GameVersion) -> Cannons:
        cannnons = {}
        if gv < 80200:
            selected_parts = [[0, 0, 0]]
        else:
            if gv > 90699:
                total_selected_parts = 0
            else:
                total_selected_parts = 10

            selected_parts = [[0, 0, 0] for _ in range(total_selected_parts)]
        return Cannons(cannnons, selected_parts)

    @staticmethod
    def read(stream: core.Data, gv: core.GameVersion) -> Cannons:
        total = stream.read_int()
        cannons: dict[int, Cannon] = {}
        for _ in range(total):
            cannon_id = stream.read_int()
            cannon = Cannon.read(stream)
            cannons[cannon_id] = cannon
        if gv < 80200:
            selected_parts = [stream.read_int_list(length=3)]
        else:
            if gv > 90699:
                total_selected_parts = stream.read_byte()
            else:
                total_selected_parts = 10

            selected_parts: list[list[int]] = []
            for _ in range(total_selected_parts):
                selected_parts.append(stream.read_byte_list(length=3))

        return Cannons(cannons, selected_parts)

    def write(self, stream: core.Data, gv: core.GameVersion):
        stream.write_int(len(self.cannons))
        for cannon_id, cannon in self.cannons.items():
            stream.write_int(cannon_id)
            cannon.write(stream)
        if gv < 80200:
            stream.write_int_list(self.selected_parts[0], write_length=False, length=3)
        else:
            if gv > 90699:
                stream.write_byte(len(self.selected_parts))

            for part in self.selected_parts:
                stream.write_byte_list(part, write_length=False, length=3)

    def serialize(self) -> dict[str, Any]:
        return {
            "cannons": {
                cannon_id: cannon.serialize()
                for cannon_id, cannon in self.cannons.items()
            },
            "selected_parts": self.selected_parts,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Cannons:
        return Cannons(
            {
                cannon_id: Cannon.deserialize(cannon)
                for cannon_id, cannon in data.get("cannons", {}).items()
            },
            data.get("selected_parts", []),
        )

    def __repr__(self):
        return f"Cannons({self.cannons}, {self.selected_parts})"

    def __str__(self):
        return f"Cannons({self.cannons}, {self.selected_parts})"


class Ototo:
    def __init__(
        self,
        base_materials: core.BaseMaterials,
        game_version: core.GameVersion | None = None,
    ):
        self.base_materials = base_materials
        self.remaining_seconds = 0.0
        self.return_flag = False
        self.improve_id = 0
        self.engineers = 0
        self.cannons = Cannons.init(game_version) if game_version else None

    @staticmethod
    def init(game_version: core.GameVersion) -> Ototo:
        return Ototo(core.BaseMaterials.init(), game_version)

    @staticmethod
    def read(stream: core.Data) -> Ototo:
        bm = core.BaseMaterials.read(stream)
        return Ototo(bm)

    def write(self, stream: core.Data):
        self.base_materials.write(stream)

    def read_2(self, stream: core.Data, gv: core.GameVersion):
        self.remaining_seconds = stream.read_double()
        self.return_flag = stream.read_bool()
        self.improve_id = stream.read_int()
        self.engineers = stream.read_int()
        self.cannons = Cannons.read(stream, gv)

    def write_2(self, stream: core.Data, gv: core.GameVersion):
        stream.write_double(self.remaining_seconds)
        stream.write_bool(self.return_flag)
        stream.write_int(self.improve_id)
        stream.write_int(self.engineers)
        if self.cannons is None:
            Cannons.init(gv).write(stream, gv)
        else:
            self.cannons.write(stream, gv)

    def serialize(self) -> dict[str, Any]:
        return {
            "base_materials": self.base_materials.serialize(),
            "remaining_seconds": self.remaining_seconds,
            "return_flag": self.return_flag,
            "improve_id": self.improve_id,
            "engineers": self.engineers,
            "cannons": self.cannons.serialize() if self.cannons else None,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Ototo:
        ototo = Ototo(core.BaseMaterials.deserialize(data.get("base_materials", [])))
        ototo.remaining_seconds = data.get("remaining_seconds", 0.0)
        ototo.return_flag = data.get("return_flag", False)
        ototo.improve_id = data.get("improve_id", 0)
        ototo.engineers = data.get("engineers", 0)
        ototo.cannons = Cannons.deserialize(data.get("cannons", {}))
        return ototo

    def __repr__(self):
        return f"Ototo({self.base_materials}, {self.remaining_seconds}, {self.return_flag}, {self.improve_id}, {self.engineers}, {self.cannons})"

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def get_max_engineers(save_file: core.SaveFile) -> int:
        file = core.core_data.get_game_data_getter(save_file).download(
            "DataLocal", "CastleCustomLimit.csv"
        )
        if file is None:
            return 5
        csv = core.CSV(file)
        return csv.lines[0][0].to_int()

    def edit_engineers(self, save_file: core.SaveFile):
        name = core.core_data.get_gatya_item_names(save_file).get_name(92)
        if name is None:
            name = "engineers"
            localized_item = True
        else:
            localized_item = False
        self.engineers = dialog_creator.SingleEditor(
            name,
            self.engineers,
            Ototo.get_max_engineers(save_file),
            localized_item=localized_item,
        ).edit()

    def display_current_cannons(self, save_file: core.SaveFile) -> list[str] | None:
        descriptions = CannonDescriptions(save_file)
        recipe_unlocks = CastleRecipeUnlock(save_file)

        color.ColoredText.localize("current_cannon_stats")

        if self.cannons is None:
            self.cannons = Cannons.init(save_file.game_version)

        names: list[str] = []
        longest_part_name = descriptions.get_longest_longest_part_name()
        if longest_part_name is None:
            return None
        longest_part_name = len(longest_part_name)

        for cannon_id, cannon in self.cannons.cannons.items():
            description = descriptions.get_cannon_description(cannon_id)
            if description is None:
                continue
            recipe_unlock = recipe_unlocks.get_recipe_unlock(cannon_id)
            if recipe_unlock is None:
                continue
            cannon_name = description.get_cannon_name()
            names.append(cannon_name)
            text = cannon_name
            if cannon_id != 0:
                cannon_name_length = len(cannon_name) - 10
                buffer = " " * (longest_part_name - cannon_name_length)
                text += core.core_data.local_manager.get_key(
                    "development",
                    development=Ototo.get_stage_name(cannon.development),
                    escape=False,
                    buffer=buffer,
                )

            for part_id, level in enumerate(cannon.levels):
                if part_id == 0:
                    level += 1

                text += "\n"
                text += "        "
                buffer = " " * (
                    longest_part_name - len(description.get_part_name(part_id)) + 2
                )
                name = description.get_part_name(part_id)
                text += core.core_data.local_manager.get_key(
                    "cannon_part", name=name, level=level, buffer=buffer
                )

            text += "\n"

            color.ColoredText.localize("cannon_stats", parts=text, escape=False)

        return names

    def edit_cannon(self, save_file: core.SaveFile):
        if self.cannons is None:
            self.cannons = Cannons.init(save_file.game_version)

        names = self.display_current_cannons(save_file)
        if names is None:
            return

        cannon_ids, all_at_once = dialog_creator.ChoiceInput.from_reduced(
            names, dialog="select_cannon"
        ).multiple_choice()
        if cannon_ids is None:
            return

        if len(cannon_ids) > 1 and not all_at_once:
            choice = dialog_creator.ChoiceInput.from_reduced(
                ["individual", "edit_all_at_once"],
                dialog="cannon_edit_type",
                single_choice=True,
            ).single_choice()
            if choice is None:
                return
            choice -= 1
            if choice == 0:
                all_at_once = False
            else:
                all_at_once = True

        if len(cannon_ids) > 1 or (len(cannon_ids) == 1 and cannon_ids[0] != 0):
            choice = dialog_creator.ChoiceInput.from_reduced(
                ["development_o", "level_o"],
                dialog="cannon_dev_level_q",
                single_choice=True,
            ).single_choice()
            if choice is None:
                return
            choice -= 1
        else:
            choice = 1
        if choice == 0:
            self.edit_cannon_development(save_file, all_at_once, cannon_ids)
        elif choice == 1:
            self.edit_cannon_level(save_file, all_at_once, cannon_ids)

        color.ColoredText.localize("cannon_success")

        self.display_current_cannons(save_file)

    def select_development(self) -> int | None:
        return dialog_creator.ChoiceInput.from_reduced(
            ["none", "foundation", "style", "effect"],
            dialog="select_development",
            single_choice=True,
        ).single_choice()

    def edit_cannon_development(
        self, save_file: core.SaveFile, all_at_once: bool, cannon_ids: list[int]
    ):
        if self.cannons is None:
            self.cannons = Cannons.init(save_file.game_version)
        if all_at_once:
            development = self.select_development()
            if development is None:
                return
            for cannon_id in cannon_ids:
                if cannon_id == 0:
                    continue
                self.cannons.cannons[cannon_id].development = development - 1
        else:
            for cannon_id in cannon_ids:
                if cannon_id == 0:
                    continue
                cannon_description = CannonDescriptions(
                    save_file
                ).get_cannon_description(cannon_id)
                if cannon_description is None:
                    continue
                current_development = self.cannons.cannons[cannon_id].development

                color.ColoredText.localize(
                    "selected_cannon_stage",
                    name=cannon_description.get_cannon_name(),
                    stage=Ototo.get_stage_name(current_development),
                    escape=False,
                )
                development = self.select_development()
                if development is None:
                    return
                self.cannons.cannons[cannon_id].development = development - 1

    def edit_cannon_level(
        self, save_file: core.SaveFile, all_at_once: bool, cannon_ids: list[int]
    ):
        if self.cannons is None:
            self.cannons = Cannons.init(save_file.game_version)
        cannon_descriptions = CannonDescriptions(save_file)
        cannon_recipe = CastleRecipeUnlock(save_file)
        if all_at_once:
            max_part_level_0 = cannon_recipe.get_max_part_level(0)
            max_part_level_1 = cannon_recipe.get_max_part_level(1)
            max_part_level_2 = cannon_recipe.get_max_part_level(2)
            if (
                max_part_level_0 is None
                or max_part_level_1 is None
                or max_part_level_2 is None
            ):
                return
            levels = dialog_creator.MultiEditor.from_reduced(
                "cannon_level",
                ["effect", "improved_foundation", "improved_style"],
                None,
                max_values=[
                    max_part_level_0,
                    max_part_level_1,
                    max_part_level_2,
                ],
                group_name_localized=True,
                items_localized=True,
            ).edit()
            if not levels:
                return
            for cannon_id in cannon_ids:
                cannon = self.get_cannon(cannon_id)
                if cannon is None:
                    continue
                cannon.development = max(cannon.development, 3)

                for part_id, level in enumerate(levels):
                    if part_id == 0:
                        level -= 1
                    max_level = cannon_recipe.get_max_level(cannon_id, part_id)
                    if max_level is None:
                        continue
                    if part_id >= len(cannon.levels):
                        break
                    cannon.levels[part_id] = min(level, max_level)
        else:
            for cannon_id in cannon_ids:
                cannon = self.get_cannon(cannon_id)
                if cannon is None:
                    continue
                cannon.development = max(cannon.development, 3)

                cannon_desc = cannon_descriptions.get_cannon_description(cannon_id)
                if cannon_desc is None:
                    continue
                levels = cannon.levels
                levels[0] += 1
                names = ["effect", "improved_foundation", "improved_style"]
                if cannon_id == 0:
                    names = ["effect"]
                max_part_level_0 = cannon_recipe.get_max_part_level(0)
                max_part_level_1 = cannon_recipe.get_max_part_level(1)
                max_part_level_2 = cannon_recipe.get_max_part_level(2)
                if (
                    max_part_level_0 is None
                    or max_part_level_1 is None
                    or max_part_level_2 is None
                ):
                    return

                levels = dialog_creator.MultiEditor.from_reduced(
                    cannon_desc.get_cannon_name(),
                    names,
                    levels,
                    max_values=[
                        max_part_level_0,
                        max_part_level_1,
                        max_part_level_2,
                    ],
                    items_localized=True,
                ).edit()
                for part_id, level in enumerate(levels):
                    if part_id == 0:
                        level -= 1
                    cannon.levels[part_id] = level

    def get_cannon(self, cannon_id: int) -> Cannon | None:
        if self.cannons is None:
            return None
        return self.cannons.cannons.get(cannon_id, None)

    @staticmethod
    def get_stage_name(development: int) -> str:
        if development == 0:
            return core.core_data.local_manager.get_key("none")
        if development == 1:
            return core.core_data.local_manager.get_key("foundation")
        if development == 2:
            return core.core_data.local_manager.get_key("style")
        if development == 3:
            return core.core_data.local_manager.get_key("effect")
        return core.core_data.local_manager.get_key("unknown_stage", stage=development)


def edit_cannon(save_file: core.SaveFile):
    save_file.ototo.edit_cannon(save_file)



from __future__ import annotations
from typing import Any
from bcps import core
from bcps.cli import color


class OfficerPass:
    def __init__(self, play_time: int):
        self.play_time = play_time
        self.gold_pass = core.NyankoClub.init()
        self.cat_id = 0
        self.cat_form = 0

    @staticmethod
    def init() -> OfficerPass:
        return OfficerPass(0)

    @staticmethod
    def read(data: core.Data) -> OfficerPass:
        play_time = data.read_int()
        return OfficerPass(play_time)

    def write(self, data: core.Data):
        if self.play_time > 2**31 - 1:
            self.play_time = 2**31 - 1
        data.write_int(self.play_time)

    def read_gold_pass(self, data: core.Data, gv: core.GameVersion):
        self.gold_pass = core.NyankoClub.read(data, gv)

    def write_gold_pass(self, data: core.Data, gv: core.GameVersion):
        self.gold_pass.write(data, gv)

    def read_cat_data(self, data: core.Data):
        self.cat_id = data.read_short()
        self.cat_form = data.read_short()

    def write_cat_data(self, data: core.Data):
        data.write_short(self.cat_id)
        data.write_short(self.cat_form)

    def serialize(self) -> dict[str, Any]:
        return {
            "play_time": self.play_time,
            "gold_pass": self.gold_pass.serialize(),
            "cat_id": self.cat_id,
            "cat_form": self.cat_form,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> OfficerPass:
        officer_pass = OfficerPass(
            data.get("play_time", 0),
        )
        officer_pass.gold_pass = core.NyankoClub.deserialize(data.get("gold_pass", {}))
        officer_pass.cat_id = data.get("cat_id", 0)
        officer_pass.cat_form = data.get("cat_form", 0)
        return officer_pass

    def __repr__(self):
        return f"OfficerPass({self.play_time}, {self.gold_pass}, {self.cat_id}, {self.cat_form})"

    def __str__(self):
        return self.__repr__()

    def reset(self, save_file: core.SaveFile):
        self.cat_id = 0
        self.cat_form = 0
        self.play_time = 0
        self.gold_pass.remove_gold_pass(save_file)

    @staticmethod
    def fix_crash(save_file: core.SaveFile):
        officer_pass = save_file.officer_pass
        officer_pass.reset(save_file)

        color.ColoredText.localize("officer_pass_fixed")



from __future__ import annotations
from bcps import core
from bcps.cli import color, dialog_creator
from typing import Union

ChaptersType = Union[
    "core.EventChapters",
    "core.GauntletChapters",
    "core.LegendQuestChapters",
    "core.ZeroLegendsChapters",
    "core.Chapters",
]


def unclear_stage(
    chapters: ChaptersType,
    map: int,
    star: int,
    stage: int,
    type: int | None = None,
):
    if isinstance(chapters, core.EventChapters):
        if type is None:
            raise ValueError("Type must be specified for EventChapters!")
        chapters.unclear_stage(type, map, star, stage)
    else:
        chapters.unclear_stage(map, star, stage)


def clear_stage(
    chapters: ChaptersType,
    map: int,
    star: int,
    stage: int,
    clear_amount: int = 1,
    overwrite_clear_progress: bool = False,
    type: int | None = None,
):
    if isinstance(chapters, core.EventChapters):
        if type is None:
            raise ValueError("Type must be specified for EventChapters!")
        chapters.clear_stage(
            type, map, star, stage, clear_amount, overwrite_clear_progress
        )
    else:
        chapters.clear_stage(map, star, stage, clear_amount, overwrite_clear_progress)


def unclear_rest(
    chapters: ChaptersType,
    stages: list[int],
    stars: int,
    id: int,
    type: int | None = None,
):
    if isinstance(chapters, core.EventChapters):
        if type is None:
            raise ValueError("Type must be specified for EventChapters!")
        chapters.unclear_rest(stages, stars, id, type)
    else:
        chapters.unclear_rest(stages, stars, id)


def get_total_stars(chapters: ChaptersType, id: int, type: int | None = None):
    if isinstance(chapters, core.EventChapters):
        if type is None:
            raise ValueError("Type must be specified for EventChapters!")
        total_stars = chapters.get_total_stars(type, id)
    else:
        total_stars = chapters.get_total_stars(id)

    return total_stars


def get_total_stages(
    chapters: ChaptersType, id: int, star: int, type: int | None = None
):
    if isinstance(chapters, core.EventChapters):
        if type is None:
            raise ValueError("Type must be specified for EventChapters!")
        total_stars = chapters.get_total_stages(type, id, star)
    else:
        total_stars = chapters.get_total_stages(id, star)

    return total_stars


def edit_chapters(
    save_file: core.SaveFile,
    chapters: ChaptersType,
    letter_code: str,
    type: int | None = None,
):
    map_names = core.MapNames(save_file, letter_code)
    names = map_names.map_names

    choice = dialog_creator.ChoiceInput.from_reduced(
        ["clear_stages", "unclear_stages"], dialog="clear_unclear_q", single_choice=True
    ).single_choice()
    if choice is None:
        return

    choice -= 1

    clear = choice == 0

    choice -= 1

    if clear:
        clear_txt = "clear"
        star_prompt = "custom_star_count_per_chapter"
    else:
        clear_txt = "unclear"
        star_prompt = "custom_star_count_per_chapter_unclear"

    map_choices = core.EventChapters.select_map_names(names)
    if not map_choices:
        return
    clear_type_choice = dialog_creator.ChoiceInput.from_reduced(
        [f"{clear_txt}_whole_chapters", f"{clear_txt}_specific_stages"],
        dialog=f"select_{clear_txt}_type",
        single_choice=True,
    ).single_choice()
    if clear_type_choice is None:
        return
    clear_type_choice -= 1

    if len(map_choices) > 1:
        stars_type_choice = dialog_creator.YesNoInput().get_input_once(
            "custom_star_count_per_chapter_yn"
        )
    else:
        stars_type_choice = False

    if stars_type_choice is None:
        return

    if clear:
        modify_clear_amounts = dialog_creator.YesNoInput().get_input_once(
            "modify_clear_amounts"
        )
    else:
        modify_clear_amounts = False

    if modify_clear_amounts is None:
        return

    clear_amount = 1
    clear_amount_type = -1
    if modify_clear_amounts:
        options = ["clear_amount_chapter", "clear_amount_all"]
        if clear_type_choice == 1:
            options.append("clear_amount_stages")
        clear_amount_type = dialog_creator.ChoiceInput.from_reduced(
            options, dialog="select_clear_amount_type", single_choice=True
        ).single_choice()
        if clear_amount_type is None:
            return
        clear_amount_type -= 1
        if clear_amount_type == 1:
            clear_amount = core.EventChapters.ask_clear_amount()
            if clear_amount is None:
                return

    if not stars_type_choice:
        stars = core.EventChapters.ask_stars(
            get_total_stars(chapters, 0, type),
            prompt=star_prompt,
        )
        if stars is None:
            return
    else:
        stars = 0

    for id in map_choices:
        map_name = names[id]
        stage_names = map_names.stage_names.get(id)
        stage_names = [
            stage_name
            for stage_name in stage_names or []
            if stage_name and stage_name != "＠"
        ]
        total_stages = len(stage_names)
        if isinstance(chapters, core.EventChapters):
            if type is None:
                raise ValueError("Type must be specified for EventChapters!")
            chapters.set_total_stages(id, type, total_stages)
        else:
            chapters.set_total_stages(id, total_stages)

        color.ColoredText.localize("current_sol_chapter", name=map_name, id=id)
        if stars_type_choice:
            stars = core.EventChapters.ask_stars(
                get_total_stars(chapters, id, type), prompt=star_prompt
            )
            if stars is None:
                return
        if clear_type_choice:
            stages = core.EventChapters.ask_stages(map_names, id)
            if stages is None:
                return
        else:
            stages = list(range(get_total_stages(chapters, id, 0, type)))

        if clear_amount_type == 0:
            clear_amount = core.EventChapters.ask_clear_amount()
            if clear_amount is None:
                return

        if not clear:
            start = stars - 1
            end = get_total_stars(chapters, id, type)
        else:
            start = 0
            end = stars
            unclear_rest(chapters, stages, stars, id, type)

        for star in range(start, end):
            if clear_amount_type == 2:
                color.ColoredText.localize("current_sol_star", star=star + 1)
            for stage in stages:
                if clear_amount_type == 2:
                    stage_name = stage_names[stage]
                    color.ColoredText.localize(
                        "current_sol_stage", name=stage_name, id=stage
                    )

                if clear_amount_type == 2:
                    clear_amount = core.EventChapters.ask_clear_amount()
                if clear_amount is None:
                    return
                if clear:
                    clear_stage(
                        chapters,
                        id,
                        star,
                        stage,
                        overwrite_clear_progress=True,
                        clear_amount=clear_amount,
                        type=type,
                    )
                else:
                    unclear_stage(chapters, id, star, stage, type)

    color.ColoredText.localize("map_chapters_edited")



from __future__ import annotations
from typing import Any
from bcps import core


class UncannyChapters:
    def __init__(self, chapters: core.Chapters, unknown: list[int]):
        self.chapters = chapters
        self.unknown = unknown

    @staticmethod
    def init() -> UncannyChapters:
        return UncannyChapters(core.Chapters.init(), [])

    @staticmethod
    def read(data: core.Data) -> UncannyChapters:
        ch = core.Chapters.read(data, read_every_time=False)
        unknown = data.read_int_list(length=len(ch.chapters))
        return UncannyChapters(ch, unknown)

    def write(self, data: core.Data):
        self.chapters.write(data, write_every_time=False)
        data.write_int_list(self.unknown, write_length=False)

    def serialize(self) -> dict[str, Any]:
        return {
            "chapters": self.chapters.serialize(),
            "unknown": self.unknown,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> UncannyChapters:
        return UncannyChapters(
            core.Chapters.deserialize(data.get("chapters", {})),
            data.get("unknown", []),
        )

    def __repr__(self):
        return f"Uncanny({self.chapters}, {self.unknown})"

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def edit_uncanny(save_file: core.SaveFile):
        uncanny = save_file.uncanny
        uncanny.chapters.edit_chapters(save_file, "NA")



from __future__ import annotations

"""ManagedItem class for bcps."""

from enum import Enum
from typing import Any
import uuid
import time
from bcps import core


class DetailType(Enum):
    """Enum for the different types of details."""

    GET = "get"
    USE = "use"


class ManagedItemType(Enum):
    """Enum for the different types of managed items."""

    CATFOOD = "catfood"
    RARE_TICKET = "rareTicket"
    PLATINUM_TICKET = "platinumTicket"
    LEGEND_TICKET = "legendTicket"


class ManagedItem:
    """Managed item for backupmetadata"""

    def __init__(
        self,
        amount: int,
        detail_type: DetailType,
        managed_item_type: ManagedItemType,
        detail_code: str = "",
        detail_created_at: int = 0,
    ):
        self.amount = amount
        self.detail_type = detail_type
        self.managed_item_type = managed_item_type
        if not detail_code:
            detail_code = str(uuid.uuid4())
        self.detail_code = detail_code
        if not detail_created_at:
            detail_created_at = int(time.time())
        self.detail_created_at = detail_created_at

    @staticmethod
    def from_change(change: int, managed_item_type: ManagedItemType) -> ManagedItem:
        """Create a managed item from a change."""
        if change > 0:
            detail_type = DetailType.GET
        else:
            detail_type = DetailType.USE
        managed_item = ManagedItem(abs(change), detail_type, managed_item_type)
        return managed_item

    def to_dict(self) -> dict[str, Any]:
        """Convert the managed item to a dictionary."""

        data = {
            "amount": self.amount,
            "detailCode": self.detail_code,
            "detailCreatedAt": self.detail_created_at,
            "detailType": self.detail_type.value,
            "managedItemType": self.managed_item_type.value,
        }
        return data

    def to_short_form(self) -> str:
        """Convert the managed item to a short form."""

        return f"{self.amount}_{self.detail_created_at}_{self.managed_item_type.value}_{self.detail_type.value}"

    @staticmethod
    def from_short_form(short_form: str) -> ManagedItem:
        values = short_form.split("_")
        try:
            amount = int(values[0])
        except (IndexError, ValueError):
            amount = 0

        try:
            detail_created_at = int(values[1])
        except (IndexError, ValueError):
            detail_created_at = 0

        try:
            managed_item_type = values[2]
        except IndexError:
            managed_item_type = ManagedItemType.CATFOOD.value

        try:
            detail_type = values[3]
        except IndexError:
            detail_type = DetailType.GET.value

        return ManagedItem(
            amount,
            DetailType(detail_type),
            ManagedItemType(managed_item_type),
            detail_created_at=detail_created_at,
        )

    def __str__(self) -> str:
        return (
            f"{self.amount} {self.managed_item_type.value} ({self.detail_type.value})"
        )

    def __repr__(self) -> str:
        return (
            f"{self.amount} {self.managed_item_type.value} ({self.detail_type.value})"
        )


class BackupMetaData:
    def __init__(
        self,
        save_file: core.SaveFile,
    ):
        self.save_file = save_file
        self.identifier = "managed_items"

    def set_managed_items(self, managed_items: list[ManagedItem]):
        self.save_file.remove_strings(self.identifier)
        for managed_item in managed_items:
            string = managed_item.to_short_form()
            self.save_file.store_string(self.identifier, string, overwrite=False)

    def get_managed_items(self) -> list[ManagedItem]:
        managed_items: list[ManagedItem] = []
        managed_items_str = self.save_file.get_strings(self.identifier)
        for managed_item_str in managed_items_str:
            managed_item = ManagedItem.from_short_form(managed_item_str)
            if managed_item.amount == 0:
                continue
            managed_items.append(managed_item)
        return managed_items

    def add_managed_item(self, managed_item: ManagedItem):
        if managed_item.amount == 0:
            return
        managed_items = self.get_managed_items()
        managed_items.append(managed_item)
        self.set_managed_items(managed_items)

    def remove_managed_items(self) -> None:
        self.save_file.remove_strings(self.identifier)

    def create(
        self, save_key: str | None = None, add_managed_items: bool = True
    ) -> str:
        """Create the backup metadata."""

        return BackupMetaData.create_static(
            self.save_file.inquiry_code,
            self.save_file.officer_pass.play_time,
            self.save_file.calculate_user_rank(),
            self.get_managed_items(),
            save_key,
            add_managed_items,
        )

    @staticmethod
    def create_static(
        iq: str,
        playtime: int,
        userrank: int,
        items: list[ManagedItem],
        save_key: str | None = None,
        add_managed_items: bool = True,
    ):
        managed_items: list[dict[str, Any]] = []
        if add_managed_items:
            for managed_item in items:
                if managed_item.amount == 0:
                    continue
                managed_items.append(managed_item.to_dict())

        managed_items_json = core.JsonFile.from_object(managed_items)
        managed_items_str = (
            managed_items_json.to_data(indent=None).to_str().replace(" ", "")
        )

        backup_metadata: dict[str, Any] = {
            "managedItemDetails": managed_items,
            "nonce": core.Random.get_hex_string(32),
            "playTime": playtime,
            "rank": userrank,
            "receiptLogIds": [],
            "signature_v1": core.NyankoSignature(
                iq, managed_items_str
            ).generate_signature_v1(),
        }
        if save_key is not None:
            backup_metadata["saveKey"] = save_key
        return (
            core.JsonFile.from_object(backup_metadata)
            .to_data(indent=None)
            .to_str()
            .replace(" ", "")
        )



from __future__ import annotations
from typing import Any
from bcps import core
from bcps.cli import color, dialog_creator, edits


class EventStage:
    def __init__(self, clear_amount: int):
        self.clear_amount = clear_amount

    @staticmethod
    def init() -> EventStage:
        return EventStage(0)

    @staticmethod
    def read(data: core.Data, is_int: bool) -> EventStage:
        if is_int:
            clear_amount = data.read_int()
        else:
            clear_amount = data.read_short()
        return EventStage(clear_amount)

    def write(self, data: core.Data, is_int: bool):
        if is_int:
            data.write_int(self.clear_amount)
        else:
            data.write_short(self.clear_amount)

    def serialize(self) -> int:
        return self.clear_amount

    @staticmethod
    def deserialize(data: int) -> EventStage:
        return EventStage(
            clear_amount=data,
        )

    def __repr__(self) -> str:
        return f"<EventStage clear_amount={self.clear_amount}>"

    def __str__(self) -> str:
        return self.__repr__()

    def clear_stage(self, clear_amount: int = 1):
        self.clear_amount = clear_amount

    def unclear_stage(self):
        self.clear_amount = 0


class EventSubChapter:
    def __init__(self, selected_stage: int, total_stages: int = 0):
        self.selected_stage = selected_stage
        self.clear_progress = 0
        self.stages = [EventStage.init() for _ in range(total_stages)]
        self.chapter_unlock_state = 0
        self.total_stages = 0

    def clear_stage(
        self,
        index: int,
        clear_amount: int = 1,
        overwrite_clear_progress: bool = False,
    ) -> bool:
        if overwrite_clear_progress:
            self.clear_progress = index + 1
        else:
            self.clear_progress = max(self.clear_progress, index + 1)
        self.stages[index].clear_stage(clear_amount)
        self.chapter_unlock_state = 3
        if index == self.total_stages - 1:
            return True
        return False

    def unclear_stage(self, index: int) -> bool:
        self.clear_progress = min(self.clear_progress, index)
        self.stages[index].unclear_stage()

        return True

    def clear_map(self, increment: bool = True) -> bool:
        self.clear_progress = len(self.stages)
        self.chapter_unlock_state = 3
        for stage in self.stages:
            if increment:
                clear_amount = stage.clear_amount + 1
            else:
                clear_amount = stage.clear_amount or 1
            stage.clear_stage(clear_amount)
        return True

    @staticmethod
    def init(total_stages: int) -> EventSubChapter:
        return EventSubChapter(0, total_stages)

    @staticmethod
    def read_selected_stage(data: core.Data, is_int: bool) -> EventSubChapter:
        if is_int:
            selected_stage = data.read_int()
        else:
            selected_stage = data.read_byte()
        return EventSubChapter(selected_stage)

    def write_selected_stage(self, data: core.Data, is_int: bool):
        if is_int:
            data.write_int(self.selected_stage)
        else:
            data.write_byte(self.selected_stage)

    def read_clear_progress(self, data: core.Data, is_int: bool):
        if is_int:
            self.clear_progress = data.read_int()
        else:
            self.clear_progress = data.read_byte()

    def write_clear_progress(self, data: core.Data, is_int: bool):
        if is_int:
            data.write_int(self.clear_progress)
        else:
            data.write_byte(self.clear_progress)

    def read_stages(self, data: core.Data, total_stages: int, is_int: bool):
        self.stages = [EventStage.read(data, is_int) for _ in range(total_stages)]

    def write_stages(self, data: core.Data, is_int: bool):
        for stage in self.stages:
            stage.write(data, is_int)

    def read_chapter_unlock_state(self, data: core.Data, is_int: bool):
        if is_int:
            self.chapter_unlock_state = data.read_int()
        else:
            self.chapter_unlock_state = data.read_byte()

    def write_chapter_unlock_state(self, data: core.Data, is_int: bool):
        if is_int:
            data.write_int(self.chapter_unlock_state)
        else:
            data.write_byte(self.chapter_unlock_state)

    def serialize(self) -> dict[str, Any]:
        return {
            "selected_stage": self.selected_stage,
            "clear_progress": self.clear_progress,
            "stages": [stage.serialize() for stage in self.stages],
            "chapter_unlock_state": self.chapter_unlock_state,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> EventSubChapter:
        sub_chapter = EventSubChapter(
            selected_stage=data.get("selected_stage", 0),
        )
        sub_chapter.clear_progress = data.get("clear_progress", 0)
        sub_chapter.stages = [
            EventStage.deserialize(stage) for stage in data.get("stages", [])
        ]
        sub_chapter.chapter_unlock_state = data.get("chapter_unlock_state", 0)
        return sub_chapter

    def __repr__(self) -> str:
        return f"<EventSubChapter selected_stage={self.selected_stage}, clear_progress={self.clear_progress}, stages={self.stages}, chapter_unlock_state={self.chapter_unlock_state}>"

    def __str__(self) -> str:
        return self.__repr__()


class EventSubChapterStars:
    def __init__(self, chapters: list[EventSubChapter]):
        self.chapters = chapters
        self.legend_restriction = 0

    def clear_stage(
        self,
        star: int,
        stage: int,
        clear_amount: int = 1,
        overwrite_clear_progress: bool = False,
    ) -> bool:
        finished = self.chapters[star].clear_stage(
            stage, clear_amount, overwrite_clear_progress
        )
        if finished:
            if star + 1 < len(self.chapters):
                self.chapters[star + 1].chapter_unlock_state = 1
        return finished

    def unclear_stage(self, star: int, stage: int):
        finished = self.chapters[star].unclear_stage(stage)
        if finished and star + 1 < len(self.chapters):
            for chapter in self.chapters[star + 1 :]:
                chapter.chapter_unlock_state = 0
        return finished

    def clear_map(self, star: int, increment: bool = True) -> bool:
        finished = self.chapters[star].clear_map(increment)
        if finished:
            if star + 1 < len(self.chapters):
                self.chapters[star + 1].chapter_unlock_state = 1
        return finished

    def clear_chapter(self, increment: bool = True) -> bool:
        for chapter in self.chapters:
            chapter.clear_map(increment)
        return True

    @staticmethod
    def init(total_stars: int) -> EventSubChapterStars:
        return EventSubChapterStars(
            [EventSubChapter.init(0) for _ in range(total_stars)]
        )

    @staticmethod
    def read_selected_stage(
        data: core.Data, total_stars: int, is_int: bool
    ) -> EventSubChapterStars:
        chapters = [
            EventSubChapter.read_selected_stage(data, is_int)
            for _ in range(total_stars)
        ]
        return EventSubChapterStars(chapters)

    def write_selected_stage(self, data: core.Data, is_int: bool):
        for chapter in self.chapters:
            chapter.write_selected_stage(data, is_int)

    def read_clear_progress(self, data: core.Data, is_int: bool):
        for chapter in self.chapters:
            chapter.read_clear_progress(data, is_int)

    def write_clear_progress(self, data: core.Data, is_int: bool):
        for chapter in self.chapters:
            chapter.write_clear_progress(data, is_int)

    def read_stages(self, data: core.Data, total_stages: int, is_int: bool):
        for _ in range(total_stages):
            for chapter in self.chapters:
                chapter.stages.append(EventStage.read(data, is_int))
                # chapter.read_stages(data, total_stages, is_int)

    def write_stages(self, data: core.Data, is_int: bool):
        for i in range(len(self.chapters[0].stages)):
            for chapter in self.chapters:
                chapter.stages[i].write(data, is_int)
                # chapter.write_stages(data, is_int)

    def read_chapter_unlock_state(self, data: core.Data, is_int: bool):
        for chapter in self.chapters:
            chapter.read_chapter_unlock_state(data, is_int)

    def write_chapter_unlock_state(self, data: core.Data, is_int: bool):
        for chapter in self.chapters:
            chapter.write_chapter_unlock_state(data, is_int)

    def read_legend_restrictions(self, data: core.Data):
        self.legend_restriction = data.read_int()

    def write_legend_restrictions(self, data: core.Data):
        data.write_int(self.legend_restriction)

    def serialize(self) -> dict[str, Any]:
        return {
            "chapters": [chapter.serialize() for chapter in self.chapters],
            "legend_restriction": self.legend_restriction,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> EventSubChapterStars:
        chapters = [
            EventSubChapter.deserialize(chapter) for chapter in data.get("chapters", [])
        ]
        chapter = EventSubChapterStars(chapters)
        chapter.legend_restriction = data.get("legend_restriction", 0)
        return chapter

    def __repr__(self) -> str:
        return f"<EventSubChapterStars chapters={self.chapters}, legend_restriction={self.legend_restriction}>"

    def __str__(self) -> str:
        return self.__repr__()


class EventChapterGroup:
    def __init__(self, chapters: list[EventSubChapterStars]):
        self.chapters = chapters

    def clear_stage(
        self,
        map: int,
        star: int,
        stage: int,
        clear_amount: int = 1,
        overwrite_clear_progress: bool = False,
    ):
        finished = self.chapters[map].clear_stage(
            star, stage, clear_amount, overwrite_clear_progress
        )
        if finished and map + 1 < len(self.chapters):
            self.chapters[map + 1].chapters[0].chapter_unlock_state = 1

    def unclear_stage(self, map: int, star: int, stage: int):
        finished = self.chapters[map].unclear_stage(star, stage)
        if finished and map + 1 < len(self.chapters) and star == 0:
            for chapter in self.chapters[map + 1].chapters:
                chapter.chapter_unlock_state = 0

    def clear_map(self, map: int, star: int, increment: bool = True):
        finished = self.chapters[map].clear_map(star, increment)
        if finished and map + 1 < len(self.chapters):
            self.chapters[map + 1].chapters[0].chapter_unlock_state = 1

    def clear_chapter(self, map: int, increment: bool = True):
        finished = self.chapters[map].clear_chapter(increment)
        if finished and map + 1 < len(self.chapters):
            self.chapters[map + 1].chapters[0].chapter_unlock_state = 1

    def clear_group(self, increment: bool = True):
        for chapter in self.chapters:
            chapter.clear_chapter(increment)

    @staticmethod
    def init(total_subchapters: int, total_stars: int) -> EventChapterGroup:
        return EventChapterGroup(
            [EventSubChapterStars.init(total_stars) for _ in range(total_subchapters)]
        )

    @staticmethod
    def read_selected_stage(
        data: core.Data, total_subchapters: int, total_stars: int, is_int: bool
    ) -> EventChapterGroup:
        chapters = [
            EventSubChapterStars.read_selected_stage(data, total_stars, is_int)
            for _ in range(total_subchapters)
        ]
        return EventChapterGroup(chapters)

    def write_selected_stage(self, data: core.Data, is_int: bool):
        for chapter in self.chapters:
            chapter.write_selected_stage(data, is_int)

    def read_clear_progress(self, data: core.Data, is_int: bool):
        for chapter in self.chapters:
            chapter.read_clear_progress(data, is_int)

    def write_clear_progress(self, data: core.Data, is_int: bool):
        for chapter in self.chapters:
            chapter.write_clear_progress(data, is_int)

    def read_stages(self, data: core.Data, total_stages: int, is_int: bool):
        for chapter in self.chapters:
            chapter.read_stages(data, total_stages, is_int)

    def write_stages(self, data: core.Data, is_int: bool):
        for chapter in self.chapters:
            chapter.write_stages(data, is_int)

    def read_chapter_unlock_state(self, data: core.Data, is_int: bool):
        for chapter in self.chapters:
            chapter.read_chapter_unlock_state(data, is_int)

    def write_chapter_unlock_state(self, data: core.Data, is_int: bool):
        for chapter in self.chapters:
            chapter.write_chapter_unlock_state(data, is_int)

    def read_legend_restrictions(self, data: core.Data):
        for chapter in self.chapters:
            chapter.read_legend_restrictions(data)

    def write_legend_restrictions(self, data: core.Data):
        for chapter in self.chapters:
            chapter.write_legend_restrictions(data)

    def serialize(self) -> list[dict[str, Any]]:
        return [chapter.serialize() for chapter in self.chapters]

    @staticmethod
    def deserialize(data: list[dict[str, Any]]) -> EventChapterGroup:
        chapters = [EventSubChapterStars.deserialize(chapter) for chapter in data]
        return EventChapterGroup(chapters)

    def __repr__(self) -> str:
        return f"<EventChapterGroup chapters={self.chapters}>"

    def __str__(self) -> str:
        return self.__repr__()


class EventChapters:
    def __init__(self, chapters: list[EventChapterGroup]):
        self.chapters = chapters
        self.completed_one_level_in_chapter: dict[int, int] = {}
        self.displayed_cleared_limit_text: dict[int, bool] = {}
        self.event_start_dates: dict[int, int] = {}
        self.stages_reward_claimed: list[int] = []

    def clear_stage(
        self,
        type: int,
        map: int,
        star: int,
        stage: int,
        clear_amount: int = 1,
        overwrite_clear_progress: bool = False,
    ):
        self.chapters[type].clear_stage(
            map,
            star,
            stage,
            clear_amount,
            overwrite_clear_progress,
        )

    def unclear_stage(self, type: int, map: int, star: int, stage: int):
        self.chapters[type].unclear_stage(map, star, stage)

    def clear_map(self, type: int, map: int, star: int, increment: bool = True):
        self.chapters[type].clear_map(map, star, increment)

    def clear_chapter(self, type: int, map: int, increment: bool = True):
        self.chapters[type].clear_chapter(map, increment)

    def clear_group(self, type: int, increment: bool = True):
        self.chapters[type].clear_group(increment)

    @staticmethod
    def init(gv: core.GameVersion) -> EventChapters:
        if gv < 20:
            return EventChapters([])
        if gv <= 32:
            total_map_types = 3
            total_subchapters = 150
            stars_per_subchapter = 3
        elif gv <= 34:
            total_map_types = 4
            total_subchapters = 150
            stars_per_subchapter = 3
        else:
            total_map_types = 0
            total_subchapters = 0
            stars_per_subchapter = 0

        return EventChapters(
            [
                EventChapterGroup.init(total_subchapters, stars_per_subchapter)
                for _ in range(total_map_types)
            ]
        )

    @staticmethod
    def read(data: core.Data, gv: core.GameVersion) -> EventChapters:
        if gv < 20:
            return EventChapters([])
        stages_per_subchapter = 0
        if 80099 < gv:
            total_map_types = data.read_byte()
            total_subchapters = data.read_short()
            stars_per_subchapter = data.read_byte()
            stages_per_subchapter = data.read_byte()
            is_int = False
        elif gv <= 32:
            total_map_types = 3
            total_subchapters = 150
            stars_per_subchapter = 3
            is_int = True
        elif gv <= 34:
            total_map_types = 4
            total_subchapters = 150
            stars_per_subchapter = 3
            is_int = True
        else:
            total_map_types = data.read_int()
            total_subchapters = data.read_int()
            stars_per_subchapter = data.read_int()
            is_int = True
        chapters = [
            EventChapterGroup.read_selected_stage(
                data, total_subchapters, stars_per_subchapter, is_int
            )
            for _ in range(total_map_types)
        ]
        if 80099 < gv:
            is_int = False
        elif gv <= 32:
            total_map_types = 3
            total_subchapters = 150
            stars_per_subchapter = 3
            is_int = True
        elif gv <= 34:
            total_map_types = 4
            total_subchapters = 150
            stars_per_subchapter = 3
            is_int = True
        else:
            total_map_types = data.read_int()
            total_subchapters = data.read_int()
            stars_per_subchapter = data.read_int()
            is_int = True

        for chapter in chapters:
            chapter.read_clear_progress(data, is_int)

        if 80099 < gv:
            is_int = False
        elif gv <= 32:
            total_map_types = 3
            total_subchapters = 150
            stars_per_subchapter = 3
            stages_per_subchapter = 12
            is_int = True
        elif gv <= 34:
            total_map_types = 4
            total_subchapters = 150
            stars_per_subchapter = 3
            stages_per_subchapter = 12
            is_int = True
        else:
            total_map_types = data.read_int()
            total_subchapters = data.read_int()
            stages_per_subchapter = data.read_int()
            stars_per_subchapter = data.read_int()
            is_int = True

        for chapter in chapters:
            chapter.read_stages(data, stages_per_subchapter, is_int)

        if 80099 < gv:
            is_int = False
        elif gv <= 32:
            total_map_types = 3
            total_subchapters = 150
            stars_per_subchapter = 3
            is_int = True
        elif gv <= 34:
            total_map_types = 4
            total_subchapters = 150
            stars_per_subchapter = 3
            is_int = True
        else:
            total_map_types = data.read_int()
            total_subchapters = data.read_int()
            stars_per_subchapter = data.read_int()
            is_int = True

        for chapter in chapters:
            chapter.read_chapter_unlock_state(data, is_int)

        return EventChapters(chapters)

    def get_lengths(self) -> tuple[int, int, int, int]:
        total_map_types = len(self.chapters)
        try:
            total_subchapters = len(self.chapters[0].chapters)
        except IndexError:
            total_subchapters = 0

        try:
            stars_per_subchapter = len(self.chapters[0].chapters[0].chapters)
        except IndexError:
            stars_per_subchapter = 0

        try:
            stages_per_subchapter = len(self.chapters[0].chapters[0].chapters[0].stages)
        except IndexError:
            stages_per_subchapter = 0
        return (
            total_map_types,
            total_subchapters,
            stars_per_subchapter,
            stages_per_subchapter,
        )

    def write(self, data: core.Data, gv: core.GameVersion):
        (
            total_map_types,
            total_subchapters,
            stars_per_subchapter,
            stages_per_subchapter,
        ) = self.get_lengths()
        if gv <= 34:
            is_int = True
        else:
            if 80099 < gv:
                data.write_byte(total_map_types)
                data.write_short(total_subchapters)
                data.write_byte(stars_per_subchapter)
                data.write_byte(stages_per_subchapter)
                is_int = False
            else:
                data.write_int(total_map_types)
                data.write_int(total_subchapters)
                data.write_int(stars_per_subchapter)
                is_int = True

        for chapter in self.chapters:
            chapter.write_selected_stage(data, is_int)

        if gv <= 34:
            is_int = True
        else:
            if 80099 < gv:
                is_int = False
            else:
                data.write_int(total_map_types)
                data.write_int(total_subchapters)
                data.write_int(stars_per_subchapter)
                is_int = True

        for chapter in self.chapters:
            chapter.write_clear_progress(data, is_int)

        if gv <= 34:
            is_int = True
        else:
            if 80099 < gv:
                is_int = False
            else:
                data.write_int(total_map_types)
                data.write_int(total_subchapters)
                data.write_int(stages_per_subchapter)
                data.write_int(stars_per_subchapter)
                is_int = True

        for chapter in self.chapters:
            chapter.write_stages(data, is_int)

        if gv <= 34:
            is_int = True
        else:
            if 80099 < gv:
                is_int = False
            else:
                data.write_int(total_map_types)
                data.write_int(total_subchapters)
                data.write_int(stars_per_subchapter)
                is_int = True

        for chapter in self.chapters:
            chapter.write_chapter_unlock_state(data, is_int)

    def read_legend_restrictions(self, data: core.Data, gv: core.GameVersion):
        if gv < 20:
            return
        if gv < 33:
            total_map_types = 3  # type: ignore
            total_subchapters = 150  # type: ignore
        elif gv < 41:
            total_map_types = 4  # type: ignore
            total_subchapters = 150  # type: ignore
        else:
            total_map_types = data.read_int()  # type: ignore
            total_subchapters = data.read_int()  # type: ignore

        for chapter in self.chapters:
            chapter.read_legend_restrictions(data)

    def write_legend_restrictions(self, data: core.Data, gv: core.GameVersion):
        if gv < 20:
            return
        if gv >= 41:
            data.write_int(len(self.chapters))
            try:
                data.write_int(len(self.chapters[0].chapters))
            except IndexError:
                data.write_int(0)

        for chapter in self.chapters:
            chapter.write_legend_restrictions(data)

    def read_dicts(self, data: core.Data):
        self.completed_one_level_in_chapter = data.read_int_int_dict()
        self.displayed_cleared_limit_text = data.read_int_bool_dict()
        self.event_start_dates = data.read_int_int_dict()
        self.stages_reward_claimed = data.read_int_list()

    def write_dicts(self, data: core.Data):
        data.write_int_int_dict(self.completed_one_level_in_chapter)
        data.write_int_bool_dict(self.displayed_cleared_limit_text)
        data.write_int_int_dict(self.event_start_dates)
        data.write_int_list(self.stages_reward_claimed)

    def serialize(self) -> dict[str, Any]:
        return {
            "chapters": [chapter.serialize() for chapter in self.chapters],
            "completed_one_level_in_chapter": self.completed_one_level_in_chapter,
            "displayed_cleared_limit_text": self.displayed_cleared_limit_text,
            "event_start_dates": self.event_start_dates,
            "stages_reward_claimed": self.stages_reward_claimed,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> EventChapters:
        chapters = [
            EventChapterGroup.deserialize(chapter)
            for chapter in data.get("chapters", [])
        ]
        ch = EventChapters(chapters)
        ch.completed_one_level_in_chapter = data.get(
            "completed_one_level_in_chapter", {}
        )
        ch.displayed_cleared_limit_text = data.get("displayed_cleared_limit_text", {})
        ch.event_start_dates = data.get("event_start_dates", {})
        ch.stages_reward_claimed = data.get("stages_reward_claimed", [])
        return ch

    def __repr__(self) -> str:
        return f"EventChapters({self.chapters}, {self.completed_one_level_in_chapter}, {self.displayed_cleared_limit_text}, {self.event_start_dates}, {self.stages_reward_claimed})"

    def __str__(self) -> str:
        return self.__repr__()

    def get_total_stars(self, type: int, map: int) -> int:
        return len(self.chapters[type].chapters[map].chapters)

    def get_total_stages(self, type: int, map: int, star: int) -> int:
        return len(self.chapters[type].chapters[map].chapters[star].stages)

    @staticmethod
    def ask_stars(
        max_stars: int, prompt: str = "custom_star_count_per_chapter"
    ) -> int | None:
        stars = dialog_creator.IntInput(min=0, max=max_stars).get_basic_input_locale(
            prompt, {"max": max_stars}
        )
        if stars is None:
            return None
        return stars

    @staticmethod
    def ask_stages(map_names: core.MapNames, chapter_id: int) -> list[int] | None:
        stage_names = map_names.stage_names.get(chapter_id)
        if stage_names is None:
            return None
        new_stage_names: list[str] = []
        for stage in stage_names:
            if stage == "＠":
                continue
            new_stage_names.append(stage)
        stage_names = new_stage_names

        dialog_creator.ListOutput(stage_names, ints=[], dialog="select_stage", localize_elements=False).display_locale()

        choices = dialog_creator.RangeInput(len(stage_names), 1).get_input_locale("stages_select", {})
        if choices is None:
            return None
        return [c -1 for c in choices]

    @staticmethod
    def ask_stages_stage_names(stage_names: list[str]) -> list[int] | None:
        new_stage_names: list[str] = []
        for stage in stage_names:
            if stage == "＠":
                continue
            new_stage_names.append(stage)
        stage_names = new_stage_names
        choice = dialog_creator.ChoiceInput.from_reduced(
            stage_names, dialog="select_stage_progress", single_choice=True
        ).single_choice()
        if choice is None:
            return None
        return list(range(choice))

    @staticmethod
    def ask_clear_amount() -> int | None:
        return dialog_creator.IntInput(
            max=core.core_data.max_value_manager.get("stage_clear_count")
        ).get_basic_input_locale("clear_amount_enter", {})

    @staticmethod
    def edit_sol_chapters(save_file: core.SaveFile):
        EventChapters.edit_chapters(save_file, 0, "N")

    @staticmethod
    def edit_event_chapters(save_file: core.SaveFile):
        EventChapters.edit_chapters(save_file, 1, "S")

    @staticmethod
    def edit_collab_chapters(save_file: core.SaveFile):
        EventChapters.edit_chapters(save_file, 2, "C")

    @staticmethod
    def select_map_names(names_dict: dict[int, str | None]) -> list[int] | None:
        map_ids: list[int] = []
        names_list: list[str] = []
        names_dict = dict(sorted(names_dict.items()))
        ids = list(names_dict.keys())
        for id, map_name in names_dict.items():
            if map_name is None:
                map_name = core.core_data.local_manager.get_key(
                    "unknown_map_name", id=id
                )
            else:
                map_name = core.core_data.local_manager.get_key(
                    "map_name", name=map_name, id=id
                )
            names_list.append(map_name)

        while True:
            dialog_creator.ListOutput(
                names_list, [], "select_map", localize_elements=False
            ).display_locale()
            if names_list:
                example_name = names_list[0]
            else:
                example_name = ""
            usr_input = (
                color.ColoredInput()
                .localize("select_map_dialog", example=example_name, escape=False)
                .lower()
                .strip()
            )
            if usr_input == "q":
                return None
            usr_ids = dialog_creator.RangeInput(max=len(names_list), min=1).parse(
                usr_input
            )
            if not usr_ids:
                found_names: list[tuple[int, str]] = []
                for i, name in enumerate(names_list):
                    if usr_input.replace(" ", "_") in name.lower().strip().replace(
                        " ", "_"
                    ):
                        true_id = ids[i]
                        found_names.append((i, name))

                if len(found_names) == 0:
                    color.ColoredText.localize("no_map_found", name=usr_input)
                elif len(found_names) == 1:
                    id = found_names[0][0]
                    true_id = ids[id]
                    if true_id not in map_ids:
                        map_ids.append(true_id)
                else:
                    selected_ids, _ = dialog_creator.ChoiceInput.from_reduced(
                        [name for _, name in found_names],
                        dialog="select_map_from_names",
                    ).multiple_choice(False)
                    if selected_ids is None:
                        continue
                    for i in selected_ids:
                        id = found_names[i][0]
                        true_id = ids[id]
                        if true_id not in map_ids:
                            map_ids.append(true_id)
            else:
                for id in usr_ids:
                    id -= 1
                    true_id = ids[id]
                    if true_id not in map_ids:
                        map_ids.append(true_id)

            color.ColoredText.localize("current_maps", maps=map_ids)

            for id in map_ids:
                name = names_dict[id]
                EventChapters.print_current_chapter(name, id)

            finished = dialog_creator.YesNoInput().get_input_once(
                "finished_selecting_maps"
            )
            if finished is None:
                return None
            if finished:
                break
        return map_ids

    @staticmethod
    def print_current_chapter(name: str | None, id: int):
        if name is None:
            name = core.core_data.local_manager.get_key("unknown_map_name", id=id)
        color.ColoredText.localize("current_sol_chapter", name=name, id=id)

    @staticmethod
    def edit_chapters(save_file: core.SaveFile, type: int, letter_code: str):
        edits.map.edit_chapters(save_file, save_file.event_stages, letter_code, type)

    def unclear_rest(
        self,
        stages: list[int],
        stars: int,
        id: int,
        type: int,
    ):
        if not stages:
            return
        for star in range(stars, self.get_total_stars(type, id)):
            for stage in range(max(stages), self.get_total_stages(type, id, star)):
                self.chapters[type].chapters[id].chapters[star].stages[
                    stage
                ].clear_amount = 0
                self.chapters[type].chapters[id].chapters[star].clear_progress = 0

    def set_total_stages(self, map: int, type: int, total_stages: int):
        for chapter in self.chapters[type].chapters[map].chapters:
            chapter.total_stages = total_stages



from __future__ import annotations
import base64
import enum
from io import BytesIO
import struct
import typing
from typing import Any, Literal
from bcps import core
import datetime


class PaddingType(enum.Enum):
    PKCS7 = enum.auto()
    ZERO = enum.auto()
    NONE = enum.auto()


class Data:
    def __init__(self, data: bytes | str | None | int | bool | Data | Any = None):
        if isinstance(data, str):
            self.data = data.encode("utf-8")
        elif isinstance(data, bytes):
            self.data = data
        elif isinstance(data, bool):
            value = 1 if data else 0
            self.data = str(value).encode("utf-8")
        elif isinstance(data, int):
            self.data = str(data).encode("utf-8")
        elif isinstance(data, Data):
            self.data = data.data
        elif data is None:
            self.data = b""
        elif hasattr(data, "__bytes__"):
            self.data = bytes(data)
        else:
            raise TypeError(
                f"data must be bytes, str, int, bool, Data, or None, not {type(data)}"
            )
        self.pos = 0
        self.set_little_endiness()
        self.buffer_enabled = False

    def __bytes__(self) -> bytes:
        return self.data

    def __buffer__(self, flags: int, /) -> memoryview:
        return memoryview(self.data)

    @staticmethod
    def from_hex(hex: str) -> Data:
        return Data(bytes.fromhex(hex))

    def enable_buffer(self):
        self.data_buffer: list[bytes] = []
        self.buffer_enabled = True

    def end_buffer(self):
        self.buffer_enabled = False
        self.data = b"".join(self.data_buffer)
        self.data_buffer = []

    def set_endiness(self, endiness: Literal["<", ">"]):
        self.endiness = endiness

    def set_little_endiness(self):
        self.set_endiness("<")

    def set_big_endiness(self):
        self.set_endiness(">")

    def is_empty(self) -> bool:
        return len(self.data) == 0

    def to_file(self, path: core.Path):
        with open(path.path, "wb") as f:
            f.write(self.data)

    def copy(self) -> Data:
        return Data(self.data)

    @staticmethod
    def from_file(path: core.Path) -> Data:
        with open(path.path, "rb") as f:
            return Data(f.read())

    def set_pos(self, pos: int):
        if pos < 0:
            pos = len(self.data) + pos
        self.pos = pos

    def reset_pos(self):
        self.pos = 0

    def clear(self):
        self.data = b""
        self.pos = 0

    def get_pos(self) -> int:
        return self.pos

    def to_hex(self) -> str:
        return self.data.hex()

    def __len__(self) -> int:
        return len(self.data)

    def __add__(self, other: Data) -> Data:
        return Data(self.data + other.data)

    @typing.overload
    def __getitem__(self, key: int) -> int:
        pass

    @typing.overload
    def __getitem__(self, key: slice) -> Data:
        pass

    def __getitem__(self, key: int | slice) -> int | Data:
        if isinstance(key, int):
            return self.data[key]
        elif isinstance(key, slice):  # type: ignore
            return Data(self.data[key])
        else:
            raise TypeError("key must be int or slice")

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Data):
            return self.data == other.data
        else:
            return False

    def get_bytes(self) -> bytes:
        return self.data

    def read_bytes(self, length: int) -> bytes:
        result = self.data[self.pos : self.pos + length]
        self.pos += length
        return result

    def read_to_end(self, remaining_data: int = 0) -> bytes:
        length = len(self.data) - self.pos - remaining_data
        return self.read_bytes(length)

    def read_int(self) -> int:
        result = struct.unpack(f"{self.endiness}i", self.read_bytes(4))[0]
        return result

    def read_variable_length_int(self) -> int:
        i = 0
        for _ in range(4):
            i3 = i << 7
            read = self.read_ubyte()
            i = i3 | (read & 0x7F)
            if read & 0x80 == 0:
                return i
        return i

    def write_variable_length_int(self, value: int):
        value = int(value)
        i2 = 0
        i3 = 0
        while value >= 128:
            i2 |= ((value & 0x7F) | 0x8000) << (i3 * 8)
            i3 += 1
            value >>= 7
        i4 = i2 | (value << (i3 * 8))
        i5 = i3 + 1
        for i6 in range(i5):
            self.write_ubyte((i4 >> (((i5 - i6) - 1) * 8)) & 0xFF)

    def read_int_list(self, length: int | None = None) -> list[int]:
        if length is None:
            length = self.read_int()
        result: list[int] = []
        for _ in range(length):
            result.append(self.read_int())
        return result

    def read_bool_list(self, length: int | None = None) -> list[bool]:
        if length is None:
            length = self.read_int()
        result: list[bool] = []
        for _ in range(length):
            result.append(self.read_bool())
        return result

    def read_string_list(self, length: int | None = None) -> list[str]:
        if length is None:
            length = self.read_int()
        result: list[str] = []
        for _ in range(length):
            result.append(self.read_string())
        return result

    def read_byte_list(self, length: int | None = None) -> list[int]:
        if length is None:
            length = self.read_int()
        result: list[int] = []
        for _ in range(length):
            result.append(self.read_byte())
        return result

    def read_short_list(self, length: int | None = None) -> list[int]:
        if length is None:
            length = self.read_int()
        result: list[int] = []
        for _ in range(length):
            result.append(self.read_short())
        return result

    def read_uint(self) -> int:
        result = struct.unpack(f"{self.endiness}I", self.read_bytes(4))[0]
        return result

    def read_short(self) -> int:
        result = struct.unpack(f"{self.endiness}h", self.read_bytes(2))[0]
        return result

    def read_ushort(self) -> int:
        result = struct.unpack(f"{self.endiness}H", self.read_bytes(2))[0]
        return result

    def read_byte(self) -> int:
        result = struct.unpack(f"{self.endiness}b", self.read_bytes(1))[0]
        return result

    def read_ubyte(self) -> int:
        result = struct.unpack(f"{self.endiness}B", self.read_bytes(1))[0]
        return result

    def read_float(self) -> float:
        result = struct.unpack(f"{self.endiness}f", self.read_bytes(4))[0]
        return result

    def read_double(self) -> float:
        result = struct.unpack(f"{self.endiness}d", self.read_bytes(8))[0]
        return result

    def read_string(self, length: int | None = None) -> str:
        if length is None:
            length = self.read_int()
        result = self.read_bytes(length).decode("utf-8")
        return result

    def read_utf8_string_by_char_length(self, length: int | None = None) -> str:
        if length is None:
            length = self.read_int()
        if length == 0:
            return ""
        result_bytes = b""
        result_str = ""
        while True:
            byte = self.read_bytes(1)[0]
            result_bytes += bytes([byte])
            try:
                result_str = result_bytes.decode("utf-8")
            except UnicodeDecodeError:
                continue
            if len(result_str) == length:
                break
        return result_str

    def read_long(self) -> int:
        result = struct.unpack(f"{self.endiness}q", self.read_bytes(8))[0]
        return result

    def read_ulong(self) -> int:
        result = struct.unpack(f"{self.endiness}Q", self.read_bytes(8))[0]
        return result

    def read_date(self):
        year = self.read_int()
        month = self.read_int()
        day = self.read_int()
        hour = self.read_int()
        minute = self.read_int()
        second = self.read_int()
        return datetime.datetime(year, month, day, hour, minute, second)

    def write_date(self, date: datetime.datetime):
        self.write_int(date.year)
        self.write_int(date.month)
        self.write_int(date.day)
        self.write_int(date.hour)
        self.write_int(date.minute)
        self.write_int(date.second)

    def write_bytes(self, data: bytes):
        if self.buffer_enabled:
            self.data_buffer.append(data)
        else:
            self.data += data
        self.pos += len(data)

    def write_int(self, value: int):
        value = int(value)
        self.write_bytes(struct.pack(f"{self.endiness}i", value))

    def write_uint(self, value: int):
        value = int(value)
        self.write_bytes(struct.pack(f"{self.endiness}I", value))

    def write_short(self, value: int):
        value = int(value)
        self.write_bytes(struct.pack(f"{self.endiness}h", value))

    def write_ushort(self, value: int):
        value = int(value)
        self.write_bytes(struct.pack(f"{self.endiness}H", value))

    def write_byte(self, value: int):
        value = int(value)
        self.write_bytes(struct.pack(f"{self.endiness}b", value))

    def write_ubyte(self, value: int):
        value = int(value)
        self.write_bytes(struct.pack(f"{self.endiness}B", value))

    def write_float(self, value: float):
        self.write_bytes(struct.pack(f"{self.endiness}f", value))

    def write_double(self, value: float):
        self.write_bytes(struct.pack(f"{self.endiness}d", value))

    def write_string(self, value: str, write_length: bool = True):
        if write_length:
            self.write_int(len(value.encode("utf-8")))
        self.write_bytes(value.encode("utf-8"))

    def write_long(self, value: int):
        self.write_bytes(struct.pack(f"{self.endiness}q", value))

    def write_ulong(self, value: int):
        self.write_bytes(struct.pack(f"{self.endiness}Q", value))

    def write_list(
        self,
        value: list[Any],
        data_type: str,
        empty_value: Any = None,
        write_length: bool = True,
        length: int | None = None,
    ):
        if length is None:
            length = len(value)
        if write_length:
            self.write_int(length)
        if length > len(value):
            value += [empty_value] * (length - len(value))
        elif length < len(value):
            value = value[:length]
        for item in value:
            getattr(self, f"write_{data_type}")(item)

    def write_int_list(
        self,
        value: list[int],
        write_length: bool = True,
        length: int | None = None,
    ):
        self.write_list(value, "int", 0, write_length, length)

    def write_bool_list(
        self,
        value: list[bool],
        write_length: bool = True,
        length: int | None = None,
    ):
        self.write_list(value, "bool", False, write_length, length)

    def write_string_list(
        self,
        value: list[str],
        write_length: bool = True,
        length: int | None = None,
    ):
        self.write_list(value, "string", "", write_length, length)

    def write_byte_list(
        self,
        value: list[int],
        write_length: bool = True,
        length: int | None = None,
    ):
        self.write_list(value, "byte", 0, write_length, length)

    def write_short_list(
        self,
        value: list[int],
        write_length: bool = True,
        length: int | None = None,
    ):
        self.write_list(value, "short", 0, write_length, length)

    def read_bool(self) -> bool:
        return self.read_byte() != 0

    def write_bool(self, value: bool):
        self.write_byte(int(value))

    def read_int_tuple(self) -> tuple[int, int]:
        return self.read_int(), self.read_int()

    def read_int_tuple_list(self, length: int | None = None) -> list[tuple[int, int]]:
        if length is None:
            length = self.read_int()
        result: list[tuple[int, int]] = []
        for _ in range(length):
            result.append(self.read_int_tuple())
        return result

    def write_int_tuple(self, value: tuple[int, int]):
        self.write_int(value[0])
        self.write_int(value[1])

    def write_int_tuple_list(
        self,
        value: list[tuple[int, int]],
        write_length: bool = True,
        length: int | None = None,
    ):
        self.write_list(value, "int_tuple", (0, 0), write_length, length)

    def read_int_bool_dict(self, length: int | None = None) -> dict[int, bool]:
        if length is None:
            length = self.read_int()
        result: dict[int, bool] = {}
        for _ in range(length):
            key = self.read_int()
            value = self.read_bool()
            result[key] = value
        return result

    def write_int_bool_dict(self, value: dict[int, bool], write_length: bool = True):
        if write_length:
            self.write_int(len(value))
        for key, item in value.items():
            self.write_int(key)
            self.write_bool(item)

    def read_int_int_dict(self, length: int | None = None) -> dict[int, int]:
        if length is None:
            length = self.read_int()
        result: dict[int, int] = {}
        for _ in range(length):
            key = self.read_int()
            value = self.read_int()
            result[key] = value
        return result

    def write_int_int_dict(self, value: dict[int, int], write_length: bool = True):
        if write_length:
            self.write_int(len(value))
        for key, item in value.items():
            self.write_int(key)
            self.write_int(item)

    def read_int_double_dict(self, length: int | None = None) -> dict[int, float]:
        if length is None:
            length = self.read_int()
        result: dict[int, float] = {}
        for _ in range(length):
            key = self.read_int()
            value = self.read_double()
            result[key] = value
        return result

    def write_int_double_dict(self, value: dict[int, float], write_length: bool = True):
        if write_length:
            self.write_int(len(value))
        for key, item in value.items():
            self.write_int(key)
            self.write_double(item)

    def read_short_bool_dict(self, length: int | None = None) -> dict[int, bool]:
        if length is None:
            length = self.read_int()
        result: dict[int, bool] = {}
        for _ in range(length):
            key = self.read_short()
            value = self.read_bool()
            result[key] = value
        return result

    def write_short_bool_dict(self, value: dict[int, bool], write_length: bool = True):
        if write_length:
            self.write_int(len(value))
        for key, item in value.items():
            self.write_short(key)
            self.write_bool(item)

    def unpad_pkcs7(self) -> Data | None:
        try:
            pad = self.data[-1]
        except IndexError:
            return None
        if pad > len(self.data):
            return None
        if self.data[-pad:] != bytes([pad] * pad):
            return None
        return Data(self.data[:-pad])

    def split(self, separator: bytes, maxsplit: int = -1) -> list[Data]:
        data_list: list[Data] = []
        for line in self.data.split(separator, maxsplit):
            data_list.append(Data(line))
        return data_list

    def to_int(self) -> int:
        try:
            return int(self.data.decode())
        except ValueError:
            return 0

    def to_int_little(self) -> int:
        return int.from_bytes(self.data, "little")

    def to_str(self) -> str:
        try:
            return self.data.decode(encoding="utf-8")
        except UnicodeDecodeError:
            return ""

    def to_bool(self) -> bool:
        return bool(self.to_int())

    @staticmethod
    def int_list_data_list(int_list: list[int]) -> list[Data]:
        data_list: list[Data] = []
        for integer in int_list:
            data_list.append(Data(str(integer)))
        return data_list

    @staticmethod
    def string_list_data_list(string_list: list[Any]) -> list[Data]:
        data_list: list[Data] = []
        for string in string_list:
            data_list.append(Data(str(string)))
        return data_list

    @staticmethod
    def data_list_int_list(data_list: list[Data]) -> list[int]:
        int_list: list[int] = []
        for data in data_list:
            int_list.append(data.to_int())
        return int_list

    @staticmethod
    def data_list_string_list(data_list: list[Data]) -> list[str]:
        string_list: list[str] = []
        for data in data_list:
            string_list.append(data.to_str())
        return string_list

    def to_bytes(self) -> bytes:
        return self.data

    @staticmethod
    def from_many(others: list[Data], joiner: Data | None = None) -> Data:
        data_lst: list[bytes] = []
        for other in others:
            data_lst.append(other.data)
        if joiner is None:
            return Data(b"".join(data_lst))
        else:
            return Data(joiner.data.join(data_lst))

    @staticmethod
    def from_int_list(int_list: list[int], endianess: Literal["little", "big"]) -> Data:
        bytes_data = b""
        for integer in int_list:
            bytes_data += integer.to_bytes(4, endianess)
        return Data(bytes_data)

    def strip(self) -> Data:
        return Data(self.data.strip())

    def replace(self, old_data: Data, new_data: Data) -> Data:
        return Data(self.data.replace(old_data.data, new_data.data))

    def set(self, value: bytes | str | None | int | bool) -> None:
        self.data = Data(value).data

    def to_bytes_io(self) -> BytesIO:
        return BytesIO(self.data)

    def __repr__(self) -> str:
        return f"Data({self.data!r})"

    def __str__(self) -> str:
        return self.to_str()

    def to_base_64(self) -> str:
        return base64.b64encode(self.data).decode()

    @staticmethod
    def from_base_64(string: str) -> Data:
        return Data(base64.b64decode(string))

    def to_csv(self, *args: Any, **kwargs: Any) -> core.CSV:
        return core.CSV(self, *args, **kwargs)

    def search(self, search_data: Data, start: int = 0) -> int:
        return self.data.find(search_data.data, start)

    def add_line(self, line: Data | str | None | bytes | int | bool = None) -> Data:
        line = Data(line)
        self.data += line.data + b"\r\n"
        return self


class PaddedInt:
    def __init__(self, value: int, size: int):
        self.value = value
        self.size = size

    def __int__(self):
        return self.value

    def __str__(self):
        return str(self.value).zfill(self.size)

    def __repr__(self):
        return f"PaddedInt({self.value}, {self.size})"

    def to_str(self):
        return str(self)



from bcps.core.game.map import (
    story,
    event,
    item_reward_stage,
    timed_score,
    ex_stage,
    dojo,
    outbreaks,
    tower,
    challenge,
    map_reset,
    uncanny,
    legend_quest,
    gauntlets,
    enigma,
    aku,
    zero_legends,
    chapters,
    map_names,
)

__all__ = [
    "story",
    "event",
    "item_reward_stage",
    "timed_score",
    "ex_stage",
    "dojo",
    "outbreaks",
    "tower",
    "challenge",
    "map_reset",
    "uncanny",
    "legend_quest",
    "gauntlets",
    "enigma",
    "aku",
    "zero_legends",
    "chapters",
    "map_names",
]



from __future__ import annotations
from typing import Any
from bcps import core
from bcps.cli import dialog_creator, color


class Outbreak:
    def __init__(self, cleared: bool):
        self.cleared = cleared

    @staticmethod
    def init() -> Outbreak:
        return Outbreak(False)

    @staticmethod
    def read(stream: core.Data) -> Outbreak:
        cleared = stream.read_bool()
        return Outbreak(cleared)

    def write(self, stream: core.Data):
        stream.write_bool(self.cleared)

    def serialize(self) -> bool:
        return self.cleared

    @staticmethod
    def deserialize(data: bool) -> Outbreak:
        return Outbreak(data)

    def __repr__(self) -> str:
        return f"Outbreak(cleared={self.cleared!r})"

    def __str__(self) -> str:
        return f"Outbreak(cleared={self.cleared!r})"


class Chapter:
    def __init__(self, id: int, outbreaks: dict[int, Outbreak]):
        self.id = id
        self.outbreaks = outbreaks

    def get_true_id(self) -> int:
        if self.id < 3:
            return self.id
        return self.id - 1

    @staticmethod
    def init(id: int) -> Chapter:
        return Chapter(id, {})

    @staticmethod
    def read(stream: core.Data, id: int) -> Chapter:
        total = stream.read_int()
        outbreaks: dict[int, Outbreak] = {}
        for _ in range(total):
            outbreak_id = stream.read_int()
            outbreak = Outbreak.read(stream)
            outbreaks[outbreak_id] = outbreak

        return Chapter(id, outbreaks)

    def write(self, stream: core.Data):
        stream.write_int(len(self.outbreaks))
        for outbreak_id, outbreak in self.outbreaks.items():
            stream.write_int(outbreak_id)
            outbreak.write(stream)

    def serialize(self) -> dict[int, Any]:
        return {
            outbreak_id: outbreak.serialize()
            for outbreak_id, outbreak in self.outbreaks.items()
        }

    @staticmethod
    def deserialize(data: dict[int, Any], id: int) -> Chapter:
        return Chapter(
            id,
            {
                outbreak_id: Outbreak.deserialize(outbreak_data)
                for outbreak_id, outbreak_data in data.items()
            },
        )

    def __repr__(self) -> str:
        return f"Chapter(id={self.id!r}, outbreaks={self.outbreaks!r})"

    def __str__(self) -> str:
        return self.__repr__()


class Outbreaks:
    def __init__(self, chapters: dict[int, Chapter]):
        self.chapters = chapters
        self.zombie_event_remaining_time = 0.0
        self.current_outbreaks: dict[int, Chapter] = {}

    @staticmethod
    def init() -> Outbreaks:
        return Outbreaks({})

    @staticmethod
    def read_chapters(stream: core.Data) -> Outbreaks:
        total = stream.read_int()
        chapters: dict[int, Chapter] = {}
        for _ in range(total):
            chapter_id = stream.read_int()
            chapter = Chapter.read(stream, chapter_id)
            chapters[chapter_id] = chapter

        return Outbreaks(chapters)

    def write_chapters(self, stream: core.Data):
        stream.write_int(len(self.chapters))
        for chapter_id, chapter in self.chapters.items():
            stream.write_int(chapter_id)
            chapter.write(stream)

    def read_2(self, stream: core.Data):
        self.zombie_event_remaining_time = stream.read_double()

    def write_2(self, stream: core.Data):
        stream.write_double(self.zombie_event_remaining_time)

    def read_current_outbreaks(self, stream: core.Data, gv: core.GameVersion):
        if gv <= 43:
            total_chapters = stream.read_int()
            for _ in range(total_chapters):
                stream.read_int()
                total_stage = stream.read_int()
                for _ in range(total_stage):
                    stream.read_int()
                    stream.read_bool()

        total = stream.read_int()
        current_outbreaks: dict[int, Chapter] = {}
        for _ in range(total):
            chapter_id = stream.read_int()
            chapter = Chapter.read(stream, chapter_id)
            current_outbreaks[chapter_id] = chapter

        self.current_outbreaks = current_outbreaks

    def write_current_outbreaks(self, stream: core.Data, gv: core.GameVersion):
        if gv <= 43:
            stream.write_int(0)
        stream.write_int(len(self.current_outbreaks))
        for chapter_id, chapter in self.current_outbreaks.items():
            stream.write_int(chapter_id)
            chapter.write(stream)

    def serialize(self) -> dict[str, Any]:
        return {
            "chapters": {
                chapter_id: chapter.serialize()
                for chapter_id, chapter in self.chapters.items()
            },
            "zombie_event_remaining_time": self.zombie_event_remaining_time,
            "current_outbreaks": {
                chapter_id: chapter.serialize()
                for chapter_id, chapter in self.current_outbreaks.items()
            },
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Outbreaks:
        outbreaks = Outbreaks(
            {
                chapter_id: Chapter.deserialize(chapter_data, chapter_id)
                for chapter_id, chapter_data in data.get("chapters", {}).items()
            }
        )
        outbreaks.zombie_event_remaining_time = data.get(
            "zombie_event_remaining_time", 0.0
        )
        outbreaks.current_outbreaks = {
            chapter_id: Chapter.deserialize(chapter_data, chapter_id)
            for chapter_id, chapter_data in data.get("current_outbreaks", {}).items()
        }

        return outbreaks

    def __repr__(self) -> str:
        return f"Outbreaks(chapters={self.chapters!r}, zombie_event_remaining_time={self.zombie_event_remaining_time!r}, current_outbreaks={self.current_outbreaks!r})"

    def __str__(self) -> str:
        return self.__repr__()

    def get_chapter_from_true_id(self, true_id: int) -> Chapter | None:
        if true_id < 3:
            return self.chapters.get(true_id)
        return self.chapters.get(true_id + 1)

    def get_current_chapter_from_true_id(self, true_id: int) -> Chapter | None:
        if true_id < 3:
            return self.current_outbreaks.get(true_id)
        return self.current_outbreaks.get(true_id + 1)

    def get_valid_chapters(self) -> list[Chapter]:
        new_chapters: list[Chapter] = []
        for chapter_id, chapter in self.chapters.items():
            if chapter_id < 9 and len(chapter.outbreaks) > 3:
                new_chapters.append(chapter)

        return new_chapters

    def clear_outbreak(self, chapter_id: int, stage_id: int, clear: bool):
        chapter = self.get_chapter_from_true_id(chapter_id)
        if chapter is not None:
            stage = chapter.outbreaks.get(stage_id)
            if stage is not None:
                stage.cleared = clear
        if clear:
            chapter = self.get_current_chapter_from_true_id(chapter_id)
            if chapter is not None:
                stage = chapter.outbreaks.get(stage_id)
                if stage is not None:
                    stage.cleared = False

    @staticmethod
    def edit_outbreaks(save_file: core.SaveFile):
        outbreaks = save_file.outbreaks
        chapters = outbreaks.get_valid_chapters()
        if not chapters:
            color.ColoredText.localize("no_valid_outbreaks")
            return

        options = ["clear", "unclear"]
        choice = dialog_creator.ChoiceInput.from_reduced(
            options, dialog="clear_unclear_outbreaks", single_choice=True
        ).single_choice()
        if choice is None:
            return
        choice -= 1

        clear = choice == 0

        selected_ids = core.StoryChapters.select_story_chapters(
            save_file, [chapter.get_true_id() for chapter in chapters]
        )
        if not selected_ids:
            return

        choice = core.StoryChapters.get_per_chapter(selected_ids)
        if choice is None:
            return
        if choice == 0:
            for chapter_id in selected_ids:
                stages = core.StoryChapters.select_stages(save_file, chapter_id)
                if not stages:
                    continue
                for stage in stages:
                    outbreaks.clear_outbreak(chapter_id, stage, clear)
        else:
            stages = core.StoryChapters.select_stages(save_file, 0)
            if not stages:
                return
            for stage in stages:
                for chapter_id in selected_ids:
                    outbreaks.clear_outbreak(chapter_id, stage, clear)

        if clear:
            color.ColoredText.localize("clear_outbreaks_success")
        else:
            color.ColoredText.localize("unclear_outbreaks_success")



from __future__ import annotations
from typing import Any
from bcps import core


class TowerChapters:
    def __init__(self, chapters: core.Chapters):
        self.chapters = chapters
        self.item_obtain_states: list[list[bool]] = []

    @staticmethod
    def init() -> TowerChapters:
        return TowerChapters(core.Chapters.init())

    @staticmethod
    def read(data: core.Data) -> TowerChapters:
        ch = core.Chapters.read(data)
        return TowerChapters(ch)

    def write(self, data: core.Data):
        self.chapters.write(data)

    def read_item_obtain_states(self, data: core.Data):
        total_stars = data.read_int()
        total_stages = data.read_int()
        self.item_obtain_states: list[list[bool]] = []
        for _ in range(total_stars):
            self.item_obtain_states.append(data.read_bool_list(total_stages))

    def write_item_obtain_states(self, data: core.Data):
        data.write_int(len(self.item_obtain_states))
        try:
            data.write_int(len(self.item_obtain_states[0]))
        except IndexError:
            data.write_int(0)
        for item_obtain_state in self.item_obtain_states:
            data.write_bool_list(item_obtain_state, write_length=False)

    def serialize(self) -> dict[str, Any]:
        return {
            "chapters": self.chapters.serialize(),
            "item_obtain_states": self.item_obtain_states,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> TowerChapters:
        tower = TowerChapters(
            core.Chapters.deserialize(data.get("chapters", {})),
        )
        tower.item_obtain_states = data.get("item_obtain_states", [])
        return tower

    def __repr__(self):
        return f"Tower({self.chapters}, {self.item_obtain_states})"

    def __str__(self):
        return self.__repr__()

    def get_total_stars(self, chapter_id: int) -> int:
        return len(self.chapters.chapters[chapter_id].chapters)

    def get_total_stages(self, chapter_id: int, star: int) -> int:
        return len(self.chapters.chapters[chapter_id].chapters[star].stages)

    @staticmethod
    def edit_towers(save_file: core.SaveFile):
        towers = save_file.tower
        towers.chapters.edit_chapters(save_file, "V")



from __future__ import annotations
from bcps import core


class Popup:
    def __init__(self, seen: bool):
        self.seen = seen

    @staticmethod
    def init() -> Popup:
        return Popup(False)

    @staticmethod
    def read(stream: core.Data) -> Popup:
        seen = stream.read_bool()
        return Popup(seen)

    def write(self, stream: core.Data):
        stream.write_bool(self.seen)

    def serialize(self) -> bool:
        return self.seen

    @staticmethod
    def deserialize(data: bool) -> Popup:
        return Popup(data)

    def __repr__(self) -> str:
        return f"Popup(seen={self.seen!r})"

    def __str__(self) -> str:
        return self.__repr__()


class UnlockPopups:
    def __init__(self, popups: dict[int, Popup]):
        self.popups = popups

    @staticmethod
    def init() -> UnlockPopups:
        return UnlockPopups({})

    @staticmethod
    def read(stream: core.Data) -> UnlockPopups:
        total = stream.read_int()
        popups: dict[int, Popup] = {}
        for _ in range(total):
            key = stream.read_int()
            popups[key] = Popup.read(stream)
        return UnlockPopups(popups)

    def write(self, stream: core.Data):
        stream.write_int(len(self.popups))
        for key, popup in self.popups.items():
            stream.write_int(key)
            popup.write(stream)

    def serialize(self) -> dict[int, bool]:
        return {key: popup.serialize() for key, popup in self.popups.items()}

    @staticmethod
    def deserialize(data: dict[int, bool]) -> UnlockPopups:
        return UnlockPopups(
            {int(key): Popup.deserialize(popup) for key, popup in data.items()}
        )

    def __repr__(self) -> str:
        return f"Popups(popups={self.popups!r})"

    def __str__(self) -> str:
        return self.__repr__()



from __future__ import annotations

import requests

from bcps import core


class RequestHandler:
    """Handles HTTP requests."""

    def __init__(
        self,
        url: str,
        headers: dict[str, str] | None = None,
        data: core.Data | None = None,
    ):
        """Initializes a new instance of the RequestHandler class.

        Args:
            url (str): URL to request.
            headers (dict[str, str] | None, optional): Headers to send with the request. Defaults to None.
            data (core.Data | None, optional): Data to send with the request. Defaults to None.
        """
        if data is None:
            data = core.Data()
        self.url = url
        self.headers = headers
        self.data = data

    def get(self, stream: bool = False) -> requests.Response | None:
        """Sends a GET request.

        Returns:
            requests.Response: Response from the server.
        """
        try:
            return requests.get(
                self.url,
                headers=self.headers,
                timeout=core.core_data.config.get_int(
                    core.ConfigKey.MAX_REQUEST_TIMEOUT
                ),
                stream=stream,
            )
        except requests.exceptions.ConnectionError:
            return None

    def post(self) -> requests.Response | None:
        """Sends a POST request.

        Returns:
            requests.Response: Response from the server.
        """
        try:
            return requests.post(
                self.url,
                headers=self.headers,
                data=self.data.data,
                timeout=core.core_data.config.get_int(
                    core.ConfigKey.MAX_REQUEST_TIMEOUT
                ),
            )
        except requests.exceptions.ConnectionError:
            return None
